######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for :class:`DataTransformer`."""
import os.path
from tempfile import TemporaryDirectory
import unittest
from unittest.mock import MagicMock, NonCallableMagicMock
from PySide6.QtWidgets import QApplication
from spine_engine.project_item.project_item_resource import database_resource
from spine_items.data_transformer.data_transformer import DataTransformer
from spine_items.data_transformer.data_transformer_factory import DataTransformerFactory
from spine_items.data_transformer.data_transformer_specification import DataTransformerSpecification
from spine_items.data_transformer.filter_config_path import filter_config_path
from spine_items.data_transformer.item_info import ItemInfo
from spine_items.data_transformer.settings import EntityClassRenamingSettings
from spinedb_api import append_filter_config
from tests.mock_helpers import create_mock_project, create_mock_toolbox, mock_finish_project_item_construction


class TestDataTransformer(unittest.TestCase):
    def setUp(self):
        """Set up."""
        self.toolbox = create_mock_toolbox()
        factory = DataTransformerFactory()
        item_dict = {"type": "Data Transformer", "description": "", "specification": None, "x": 0, "y": 0}
        self._temp_dir = TemporaryDirectory()
        self.project = create_mock_project(self._temp_dir.name)
        self.toolbox.project.return_value = self.project
        self.transformer = factory.make_item("T", item_dict, self.toolbox, self.project)
        self._properties_widget = mock_finish_project_item_construction(factory, self.transformer, self.toolbox)

    def tearDown(self):
        self._temp_dir.cleanup()

    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_item_type(self):
        self.assertEqual(DataTransformer.item_type(), ItemInfo.item_type())

    def test_item_dict(self):
        """Tests Item dictionary creation."""
        d = self.transformer.item_dict()
        self.assertEqual(d, {"description": "", "type": "Data Transformer", "x": 0.0, "y": 0.0, "specification": ""})

    def test_notify_destination(self):
        self.transformer.logger.msg = MagicMock()
        self.transformer.logger.msg_warning = MagicMock()
        source_item = NonCallableMagicMock()
        source_item.name = "source name"
        source_item.item_type = MagicMock(return_value="Data Store")
        self.transformer.notify_destination(source_item)
        self.transformer.logger.msg.emit.assert_called_with(
            "Link established. You can now define transformations for Data Store "
            "<b>source name</b> in Data Transformer <b>T</b>."
        )

        source_item.item_type = MagicMock(return_value="Data Transformer")
        self.transformer.notify_destination(source_item)
        self.transformer.logger.msg.emit.assert_called_with(
            "Link established. You can now define additional transformations in Data Transformer <b>T</b>."
        )
        source_item.item_type = MagicMock(return_value="Data Connection")
        self.transformer.notify_destination(source_item)
        self.transformer.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a "
            f"<b>{source_item.item_type()}</b> and a <b>{self.transformer.item_type()}</b> has not been "
            "implemented yet."
        )

    def test_resources_for_direct_successors(self):
        self.assertEqual(self.transformer.resources_for_direct_successors(), [])
        provider = MagicMock()
        provider.name = "resource provider"
        db_resource = database_resource(provider.name, "sqlite:///database.sqlite")
        self.transformer.upstream_resources_updated([db_resource])
        expected_resource = database_resource(self.transformer.name, "sqlite:///database.sqlite")
        self.assertEqual(self.transformer.resources_for_direct_successors(), [expected_resource])
        settings = EntityClassRenamingSettings({})
        specification = DataTransformerSpecification("specification", settings, "test specification")
        self.transformer.do_set_specification(specification)
        config_path = filter_config_path(self.transformer.data_dir)
        filter_url = append_filter_config("sqlite:///database.sqlite", config_path)
        expected_resource = database_resource(self.transformer.name, filter_url)
        self.assertEqual(self.transformer.resources_for_direct_successors(), [expected_resource])

    def test_rename(self):
        """Tests renaming a DT."""
        self.transformer.activate()
        expected_name = "ABC"
        expected_short_name = "abc"
        self.transformer.rename(expected_name, "")
        self.assertEqual(expected_name, self.transformer.name)
        self.assertEqual(expected_name, self.transformer.get_icon().name())
        expected_data_dir = os.path.join(self.project.items_dir, expected_short_name)
        self.assertEqual(expected_data_dir, self.transformer.data_dir)


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""The ViewFactory class."""
from PySide6.QtGui import QColor
from spinetoolbox.project_item.project_item_factory import ProjectItemFactory
from .view import View
from .view_icon import ViewIcon
from .widgets.add_view_widget import AddViewWidget
from .widgets.view_properties_widget import ViewPropertiesWidget


class ViewFactory(ProjectItemFactory):
    @staticmethod
    def item_class():
        return View

    @staticmethod
    def icon():
        return ":/icons/item_icons/binoculars.svg"

    @staticmethod
    def icon_color():
        return QColor("green")

    @staticmethod
    def make_add_item_widget(toolbox, x, y, specification):
        return AddViewWidget(toolbox, x, y, specification)

    @staticmethod
    def make_icon(toolbox):
        return ViewIcon(toolbox, ViewFactory.icon(), ViewFactory.icon_color())

    @staticmethod
    def make_item(name, item_dict, toolbox, project):
        return View.from_dict(name, item_dict, toolbox, project)

    @staticmethod
    def make_properties_widget(toolbox):
        return ViewPropertiesWidget(toolbox)

    @staticmethod
    def make_specification_menu(parent, index):
        raise NotImplementedError()

    @staticmethod
    def make_specification_editor(toolbox, specification=None, item=None, **kwargs):
        """See base class."""
        raise NotImplementedError()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Functions to write export preview tables."""
import numpy
from spinedb_api.spine_io.exporters.writer import Writer


class TableWriter(Writer):
    """An export writer that writes to a Python dictionary."""

    def __init__(self):
        self._tables = {}
        self._current_table = None

    def finish_table(self):
        self._current_table = None

    def start_table(self, table_name, title_key):
        self._current_table = self._tables.setdefault(table_name, [])
        return True

    @property
    def tables(self):
        """A dictionary containing the tables."""
        return self._tables

    def write_row(self, row):
        self._current_table.append([_sanitize(cell) for cell in row])
        return True


def _sanitize(x):
    """Converts special parameter value types to strings.

    Args:
        x (Any): parameter value

    Returns:
        float or int or str: sanitized value
    """
    if isinstance(x, numpy.float64):
        return float(x)
    if not isinstance(x, (float, str, int)) and x is not None:
        return str(x)
    return x



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for DataStoreExecutable."""
from multiprocessing import Lock
import os.path
from pathlib import Path
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from spine_engine import ExecutionDirection
from spine_engine.utils.helpers import urls_equal
from spine_items.data_store.executable_item import ExecutableItem
from spine_items.utils import convert_to_sqlalchemy_url


class TestDataStoreExecutable(unittest.TestCase):
    def setUp(self):
        self._temp_dir = TemporaryDirectory()

    def tearDown(self):
        self._temp_dir.cleanup()

    def test_item_type(self):
        self.assertEqual(ExecutableItem.item_type(), "Data Store")

    def test_from_dict(self):
        name = "Output Data Store"
        item_dict = {
            "type": "Data Store",
            "description": "",
            "x": 0,
            "y": 0,
            "url": {
                "dialect": "sqlite",
                "username": "",
                "password": "",
                "host": "",
                "port": "",
                "database": {
                    "type": "path",
                    "relative": True,
                    "path": ".spinetoolbox/items/output_data_store/Data Store 2.sqlite",
                },
            },
        }
        logger = mock.MagicMock()
        with mock.patch("spine_items.data_store.executable_item.convert_to_sqlalchemy_url") as mock_convert_url:
            mock_convert_url.return_value = "database.sqlite"
            item = ExecutableItem.from_dict(item_dict, name, self._temp_dir.name, None, {}, logger)
            mock_convert_url.assert_called_once()
            self.assertIsInstance(item, ExecutableItem)
            self.assertEqual("Data Store", item.item_type())
            self.assertEqual("database.sqlite", item._url)

    def test_stop_execution(self):
        executable = ExecutableItem("name", "", self._temp_dir.name, mock.MagicMock())
        with mock.patch(
            "spine_engine.project_item.executable_item_base.ExecutableItemBase.stop_execution"
        ) as mock_stop_execution:
            executable.stop_execution()
            mock_stop_execution.assert_called_once()

    def test_execute(self):
        executable = ExecutableItem("name", "", self._temp_dir.name, mock.MagicMock())
        self.assertTrue(executable.execute([], [], Lock()))

    def test_output_resources_backward(self):
        db_file_path = Path(self._temp_dir.name, "database.sqlite")
        db_file_path.touch()
        url = convert_to_sqlalchemy_url({"dialect": "sqlite", "database": str(db_file_path)})
        executable = ExecutableItem("name", url, self._temp_dir.name, mock.MagicMock())
        with mock.patch("spine_items.data_store.executable_item.DatabaseMapping.create_engine"):
            resources = executable.output_resources(ExecutionDirection.BACKWARD)
        self.assertEqual(len(resources), 1)
        resource = resources[0]
        self.assertEqual(resource.type_, "database")
        self.assertTrue(urls_equal(resource.url, "sqlite:///" + os.path.normcase(str(db_file_path))))
        self.assertEqual(resource.label, "db_url@name")

    def test_output_resources_forward(self):
        db_file_path = Path(self._temp_dir.name, "database.sqlite")
        db_file_path.touch()
        url = convert_to_sqlalchemy_url({"dialect": "sqlite", "database": str(db_file_path)})
        executable = ExecutableItem("name", url, self._temp_dir.name, mock.MagicMock())
        with mock.patch("spine_items.data_store.executable_item.DatabaseMapping.create_engine"):
            resources = executable.output_resources(ExecutionDirection.FORWARD)
        self.assertEqual(len(resources), 1)
        resource = resources[0]
        self.assertEqual(resource.type_, "database")
        self.assertTrue(urls_equal(resource.url, "sqlite:///" + os.path.normcase(str(db_file_path))))
        self.assertEqual(resource.label, "db_url@name")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

################################################################################
## Form generated from reading UI file 'specification_editor.ui'
##
## Created by: Qt User Interface Compiler version 6.7.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QCheckBox, QComboBox,
    QFormLayout, QFrame, QHBoxLayout, QHeaderView,
    QLabel, QLayout, QLineEdit, QMainWindow,
    QPushButton, QSizePolicy, QSpacerItem, QSpinBox,
    QSplitter, QTableView, QToolButton, QTreeView,
    QVBoxLayout, QWidget)

from spinetoolbox.widgets.custom_combobox import ElidedCombobox
from spine_items import resources_icons_rc

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(1086, 780)
        MainWindow.setDockNestingEnabled(True)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout_10 = QVBoxLayout(self.centralwidget)
        self.verticalLayout_10.setObjectName(u"verticalLayout_10")
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalLayout_3.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.label = QLabel(self.centralwidget)
        self.label.setObjectName(u"label")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)

        self.horizontalLayout_3.addWidget(self.label)

        self.export_format_combo_box = QComboBox(self.centralwidget)
        self.export_format_combo_box.setObjectName(u"export_format_combo_box")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.export_format_combo_box.sizePolicy().hasHeightForWidth())
        self.export_format_combo_box.setSizePolicy(sizePolicy1)

        self.horizontalLayout_3.addWidget(self.export_format_combo_box)

        self.live_preview_check_box = QCheckBox(self.centralwidget)
        self.live_preview_check_box.setObjectName(u"live_preview_check_box")
        self.live_preview_check_box.setChecked(True)

        self.horizontalLayout_3.addWidget(self.live_preview_check_box)

        self.frame_preview = QFrame(self.centralwidget)
        self.frame_preview.setObjectName(u"frame_preview")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.frame_preview.sizePolicy().hasHeightForWidth())
        self.frame_preview.setSizePolicy(sizePolicy2)
        self.frame_preview.setFrameShape(QFrame.Shape.StyledPanel)
        self.frame_preview.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame_preview)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(3, 0, 3, 0)
        self.label_9 = QLabel(self.frame_preview)
        self.label_9.setObjectName(u"label_9")

        self.horizontalLayout.addWidget(self.label_9)

        self.database_url_combo_box = ElidedCombobox(self.frame_preview)
        self.database_url_combo_box.setObjectName(u"database_url_combo_box")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.database_url_combo_box.sizePolicy().hasHeightForWidth())
        self.database_url_combo_box.setSizePolicy(sizePolicy3)
        self.database_url_combo_box.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.database_url_combo_box.setMinimumContentsLength(16)

        self.horizontalLayout.addWidget(self.database_url_combo_box)

        self.load_url_from_fs_button = QToolButton(self.frame_preview)
        self.load_url_from_fs_button.setObjectName(u"load_url_from_fs_button")
        icon = QIcon()
        icon.addFile(u":/icons/folder-open-solid.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.load_url_from_fs_button.setIcon(icon)

        self.horizontalLayout.addWidget(self.load_url_from_fs_button)

        self.label_3 = QLabel(self.frame_preview)
        self.label_3.setObjectName(u"label_3")

        self.horizontalLayout.addWidget(self.label_3)

        self.max_preview_tables_spin_box = QSpinBox(self.frame_preview)
        self.max_preview_tables_spin_box.setObjectName(u"max_preview_tables_spin_box")
        self.max_preview_tables_spin_box.setMaximum(16777215)
        self.max_preview_tables_spin_box.setSingleStep(10)
        self.max_preview_tables_spin_box.setValue(20)

        self.horizontalLayout.addWidget(self.max_preview_tables_spin_box)

        self.label_2 = QLabel(self.frame_preview)
        self.label_2.setObjectName(u"label_2")

        self.horizontalLayout.addWidget(self.label_2)

        self.max_preview_rows_spin_box = QSpinBox(self.frame_preview)
        self.max_preview_rows_spin_box.setObjectName(u"max_preview_rows_spin_box")
        self.max_preview_rows_spin_box.setMaximum(16777215)
        self.max_preview_rows_spin_box.setSingleStep(10)
        self.max_preview_rows_spin_box.setValue(20)

        self.horizontalLayout.addWidget(self.max_preview_rows_spin_box)


        self.horizontalLayout_3.addWidget(self.frame_preview)


        self.verticalLayout_10.addLayout(self.horizontalLayout_3)

        self.splitter_3 = QSplitter(self.centralwidget)
        self.splitter_3.setObjectName(u"splitter_3")
        self.splitter_3.setOrientation(Qt.Orientation.Horizontal)
        self.splitter_2 = QSplitter(self.splitter_3)
        self.splitter_2.setObjectName(u"splitter_2")
        self.splitter_2.setOrientation(Qt.Orientation.Vertical)
        self.mapping_list_layout_widget = QWidget(self.splitter_2)
        self.mapping_list_layout_widget.setObjectName(u"mapping_list_layout_widget")
        self.verticalLayout_9 = QVBoxLayout(self.mapping_list_layout_widget)
        self.verticalLayout_9.setSpacing(0)
        self.verticalLayout_9.setObjectName(u"verticalLayout_9")
        self.frame = QFrame(self.mapping_list_layout_widget)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.StyledPanel)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_11 = QVBoxLayout(self.frame)
        self.verticalLayout_11.setSpacing(0)
        self.verticalLayout_11.setObjectName(u"verticalLayout_11")
        self.verticalLayout_11.setContentsMargins(3, 3, 3, 3)
        self.label_11 = QLabel(self.frame)
        self.label_11.setObjectName(u"label_11")

        self.verticalLayout_11.addWidget(self.label_11)


        self.verticalLayout_9.addWidget(self.frame)

        self.mappings_table = QTableView(self.mapping_list_layout_widget)
        self.mappings_table.setObjectName(u"mappings_table")
        self.mappings_table.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.mappings_table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.mappings_table.setShowGrid(False)
        self.mappings_table.verticalHeader().setVisible(False)

        self.verticalLayout_9.addWidget(self.mappings_table)

        self.splitter_2.addWidget(self.mapping_list_layout_widget)
        self.mapping_controls_layout_widget = QWidget(self.splitter_2)
        self.mapping_controls_layout_widget.setObjectName(u"mapping_controls_layout_widget")
        self.verticalLayout_8 = QVBoxLayout(self.mapping_controls_layout_widget)
        self.verticalLayout_8.setObjectName(u"verticalLayout_8")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.add_mapping_button = QPushButton(self.mapping_controls_layout_widget)
        self.add_mapping_button.setObjectName(u"add_mapping_button")

        self.horizontalLayout_2.addWidget(self.add_mapping_button)

        self.remove_mapping_button = QPushButton(self.mapping_controls_layout_widget)
        self.remove_mapping_button.setObjectName(u"remove_mapping_button")

        self.horizontalLayout_2.addWidget(self.remove_mapping_button)

        self.toggle_enabled_button = QPushButton(self.mapping_controls_layout_widget)
        self.toggle_enabled_button.setObjectName(u"toggle_enabled_button")

        self.horizontalLayout_2.addWidget(self.toggle_enabled_button)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_2)

        self.write_earlier_button = QPushButton(self.mapping_controls_layout_widget)
        self.write_earlier_button.setObjectName(u"write_earlier_button")

        self.horizontalLayout_2.addWidget(self.write_earlier_button)

        self.write_later_button = QPushButton(self.mapping_controls_layout_widget)
        self.write_later_button.setObjectName(u"write_later_button")

        self.horizontalLayout_2.addWidget(self.write_later_button)


        self.verticalLayout_8.addLayout(self.horizontalLayout_2)

        self.mapping_options_contents = QFrame(self.mapping_controls_layout_widget)
        self.mapping_options_contents.setObjectName(u"mapping_options_contents")
        self.formLayout = QFormLayout(self.mapping_options_contents)
        self.formLayout.setObjectName(u"formLayout")
        self.label_4 = QLabel(self.mapping_options_contents)
        self.label_4.setObjectName(u"label_4")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label_4)

        self.item_type_combo_box = QComboBox(self.mapping_options_contents)
        self.item_type_combo_box.addItem("")
        self.item_type_combo_box.addItem("")
        self.item_type_combo_box.addItem("")
        self.item_type_combo_box.addItem("")
        self.item_type_combo_box.addItem("")
        self.item_type_combo_box.addItem("")
        self.item_type_combo_box.addItem("")
        self.item_type_combo_box.setObjectName(u"item_type_combo_box")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.item_type_combo_box)

        self.label_8 = QLabel(self.mapping_options_contents)
        self.label_8.setObjectName(u"label_8")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.label_8)

        self.entity_dimensions_spin_box = QSpinBox(self.mapping_options_contents)
        self.entity_dimensions_spin_box.setObjectName(u"entity_dimensions_spin_box")
        self.entity_dimensions_spin_box.setMinimum(0)

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.entity_dimensions_spin_box)

        self.label_7 = QLabel(self.mapping_options_contents)
        self.label_7.setObjectName(u"label_7")

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.label_7)

        self.highlight_dimension_spin_box = QSpinBox(self.mapping_options_contents)
        self.highlight_dimension_spin_box.setObjectName(u"highlight_dimension_spin_box")
        self.highlight_dimension_spin_box.setMinimum(1)

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.highlight_dimension_spin_box)

        self.label_5 = QLabel(self.mapping_options_contents)
        self.label_5.setObjectName(u"label_5")

        self.formLayout.setWidget(4, QFormLayout.LabelRole, self.label_5)

        self.parameter_type_combo_box = QComboBox(self.mapping_options_contents)
        self.parameter_type_combo_box.addItem("")
        self.parameter_type_combo_box.addItem("")
        self.parameter_type_combo_box.addItem("")
        self.parameter_type_combo_box.setObjectName(u"parameter_type_combo_box")

        self.formLayout.setWidget(4, QFormLayout.FieldRole, self.parameter_type_combo_box)

        self.label_10 = QLabel(self.mapping_options_contents)
        self.label_10.setObjectName(u"label_10")

        self.formLayout.setWidget(5, QFormLayout.LabelRole, self.label_10)

        self.parameter_dimensions_spin_box = QSpinBox(self.mapping_options_contents)
        self.parameter_dimensions_spin_box.setObjectName(u"parameter_dimensions_spin_box")

        self.formLayout.setWidget(5, QFormLayout.FieldRole, self.parameter_dimensions_spin_box)

        self.fix_table_name_check_box = QCheckBox(self.mapping_options_contents)
        self.fix_table_name_check_box.setObjectName(u"fix_table_name_check_box")

        self.formLayout.setWidget(7, QFormLayout.LabelRole, self.fix_table_name_check_box)

        self.fix_table_name_line_edit = QLineEdit(self.mapping_options_contents)
        self.fix_table_name_line_edit.setObjectName(u"fix_table_name_line_edit")

        self.formLayout.setWidget(7, QFormLayout.FieldRole, self.fix_table_name_line_edit)

        self.always_export_header_check_box = QCheckBox(self.mapping_options_contents)
        self.always_export_header_check_box.setObjectName(u"always_export_header_check_box")

        self.formLayout.setWidget(10, QFormLayout.LabelRole, self.always_export_header_check_box)

        self.compact_button = QPushButton(self.mapping_options_contents)
        self.compact_button.setObjectName(u"compact_button")

        self.formLayout.setWidget(10, QFormLayout.FieldRole, self.compact_button)

        self.label_6 = QLabel(self.mapping_options_contents)
        self.label_6.setObjectName(u"label_6")

        self.formLayout.setWidget(6, QFormLayout.LabelRole, self.label_6)

        self.group_fn_combo_box = QComboBox(self.mapping_options_contents)
        self.group_fn_combo_box.setObjectName(u"group_fn_combo_box")

        self.formLayout.setWidget(6, QFormLayout.FieldRole, self.group_fn_combo_box)


        self.verticalLayout_8.addWidget(self.mapping_options_contents)

        self.mapping_table_view = QTableView(self.mapping_controls_layout_widget)
        self.mapping_table_view.setObjectName(u"mapping_table_view")
        self.mapping_table_view.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.mapping_table_view.horizontalHeader().setStretchLastSection(True)
        self.mapping_table_view.verticalHeader().setVisible(False)

        self.verticalLayout_8.addWidget(self.mapping_table_view)

        self.splitter_2.addWidget(self.mapping_controls_layout_widget)
        self.splitter_3.addWidget(self.splitter_2)
        self.splitter = QSplitter(self.splitter_3)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.preview_tree_view = QTreeView(self.splitter)
        self.preview_tree_view.setObjectName(u"preview_tree_view")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy4.setHorizontalStretch(1)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.preview_tree_view.sizePolicy().hasHeightForWidth())
        self.preview_tree_view.setSizePolicy(sizePolicy4)
        self.splitter.addWidget(self.preview_tree_view)
        self.preview_tree_view.header().setVisible(False)
        self.preview_table_view = QTableView(self.splitter)
        self.preview_table_view.setObjectName(u"preview_table_view")
        sizePolicy5 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy5.setHorizontalStretch(5)
        sizePolicy5.setVerticalStretch(0)
        sizePolicy5.setHeightForWidth(self.preview_table_view.sizePolicy().hasHeightForWidth())
        self.preview_table_view.setSizePolicy(sizePolicy5)
        self.splitter.addWidget(self.preview_table_view)
        self.splitter_3.addWidget(self.splitter)

        self.verticalLayout_10.addWidget(self.splitter_3)

        MainWindow.setCentralWidget(self.centralwidget)
        QWidget.setTabOrder(self.export_format_combo_box, self.live_preview_check_box)
        QWidget.setTabOrder(self.live_preview_check_box, self.database_url_combo_box)
        QWidget.setTabOrder(self.database_url_combo_box, self.load_url_from_fs_button)
        QWidget.setTabOrder(self.load_url_from_fs_button, self.max_preview_tables_spin_box)
        QWidget.setTabOrder(self.max_preview_tables_spin_box, self.max_preview_rows_spin_box)
        QWidget.setTabOrder(self.max_preview_rows_spin_box, self.mappings_table)
        QWidget.setTabOrder(self.mappings_table, self.add_mapping_button)
        QWidget.setTabOrder(self.add_mapping_button, self.remove_mapping_button)
        QWidget.setTabOrder(self.remove_mapping_button, self.toggle_enabled_button)
        QWidget.setTabOrder(self.toggle_enabled_button, self.write_earlier_button)
        QWidget.setTabOrder(self.write_earlier_button, self.write_later_button)
        QWidget.setTabOrder(self.write_later_button, self.item_type_combo_box)
        QWidget.setTabOrder(self.item_type_combo_box, self.entity_dimensions_spin_box)
        QWidget.setTabOrder(self.entity_dimensions_spin_box, self.highlight_dimension_spin_box)
        QWidget.setTabOrder(self.highlight_dimension_spin_box, self.parameter_type_combo_box)
        QWidget.setTabOrder(self.parameter_type_combo_box, self.parameter_dimensions_spin_box)
        QWidget.setTabOrder(self.parameter_dimensions_spin_box, self.group_fn_combo_box)
        QWidget.setTabOrder(self.group_fn_combo_box, self.fix_table_name_check_box)
        QWidget.setTabOrder(self.fix_table_name_check_box, self.fix_table_name_line_edit)
        QWidget.setTabOrder(self.fix_table_name_line_edit, self.always_export_header_check_box)
        QWidget.setTabOrder(self.always_export_header_check_box, self.compact_button)
        QWidget.setTabOrder(self.compact_button, self.mapping_table_view)
        QWidget.setTabOrder(self.mapping_table_view, self.preview_tree_view)
        QWidget.setTabOrder(self.preview_tree_view, self.preview_table_view)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"MainWindow", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Export format:", None))
        self.live_preview_check_box.setText(QCoreApplication.translate("MainWindow", u"Live preview", None))
        self.label_9.setText(QCoreApplication.translate("MainWindow", u"Database url:", None))
#if QT_CONFIG(tooltip)
        self.load_url_from_fs_button.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Browse file system</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.load_url_from_fs_button.setText(QCoreApplication.translate("MainWindow", u"...", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Max. tables", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Max. content rows:", None))
        self.label_11.setText(QCoreApplication.translate("MainWindow", u"Mappings", None))
        self.add_mapping_button.setText(QCoreApplication.translate("MainWindow", u"Add", None))
        self.remove_mapping_button.setText(QCoreApplication.translate("MainWindow", u"Remove", None))
#if QT_CONFIG(tooltip)
        self.toggle_enabled_button.setToolTip(QCoreApplication.translate("MainWindow", u"Enable or disable all mappings at once.", None))
#endif // QT_CONFIG(tooltip)
        self.toggle_enabled_button.setText(QCoreApplication.translate("MainWindow", u"Toggle enabled", None))
#if QT_CONFIG(tooltip)
        self.write_earlier_button.setToolTip(QCoreApplication.translate("MainWindow", u"Prioratize mapping.", None))
#endif // QT_CONFIG(tooltip)
        self.write_earlier_button.setText(QCoreApplication.translate("MainWindow", u"Write earlier", None))
#if QT_CONFIG(tooltip)
        self.write_later_button.setToolTip(QCoreApplication.translate("MainWindow", u"Deprioratize mapping.", None))
#endif // QT_CONFIG(tooltip)
        self.write_later_button.setText(QCoreApplication.translate("MainWindow", u"Write later", None))
        self.label_4.setText(QCoreApplication.translate("MainWindow", u"Type:", None))
        self.item_type_combo_box.setItemText(0, QCoreApplication.translate("MainWindow", u"Entity class", None))
        self.item_type_combo_box.setItemText(1, QCoreApplication.translate("MainWindow", u"Entity class with dimension parameter", None))
        self.item_type_combo_box.setItemText(2, QCoreApplication.translate("MainWindow", u"Entity group", None))
        self.item_type_combo_box.setItemText(3, QCoreApplication.translate("MainWindow", u"Alternative", None))
        self.item_type_combo_box.setItemText(4, QCoreApplication.translate("MainWindow", u"Scenario", None))
        self.item_type_combo_box.setItemText(5, QCoreApplication.translate("MainWindow", u"Scenario alternative", None))
        self.item_type_combo_box.setItemText(6, QCoreApplication.translate("MainWindow", u"Parameter value list", None))

        self.label_8.setText(QCoreApplication.translate("MainWindow", u"Entity dimensions:", None))
#if QT_CONFIG(tooltip)
        self.entity_dimensions_spin_box.setToolTip(QCoreApplication.translate("MainWindow", u"Number of expected relationship dimensions.", None))
#endif // QT_CONFIG(tooltip)
        self.label_7.setText(QCoreApplication.translate("MainWindow", u"Selected dimension:", None))
#if QT_CONFIG(tooltip)
        self.highlight_dimension_spin_box.setToolTip(QCoreApplication.translate("MainWindow", u"Relationship dimension used to select object parameters for export.", None))
#endif // QT_CONFIG(tooltip)
        self.label_5.setText(QCoreApplication.translate("MainWindow", u"Parameter type:", None))
        self.parameter_type_combo_box.setItemText(0, QCoreApplication.translate("MainWindow", u"Value", None))
        self.parameter_type_combo_box.setItemText(1, QCoreApplication.translate("MainWindow", u"Default value", None))
        self.parameter_type_combo_box.setItemText(2, QCoreApplication.translate("MainWindow", u"None", None))

        self.label_10.setText(QCoreApplication.translate("MainWindow", u"Parameter dimensions:", None))
#if QT_CONFIG(tooltip)
        self.parameter_dimensions_spin_box.setToolTip(QCoreApplication.translate("MainWindow", u"Number of expected parameter value dimensions.", None))
#endif // QT_CONFIG(tooltip)
        self.fix_table_name_check_box.setText(QCoreApplication.translate("MainWindow", u"Fixed table name:", None))
#if QT_CONFIG(tooltip)
        self.always_export_header_check_box.setToolTip(QCoreApplication.translate("MainWindow", u"Export header even when a table is otherwise empty.", None))
#endif // QT_CONFIG(tooltip)
        self.always_export_header_check_box.setText(QCoreApplication.translate("MainWindow", u"Always export header", None))
#if QT_CONFIG(tooltip)
        self.compact_button.setToolTip(QCoreApplication.translate("MainWindow", u"Compact mapping by removing empty columns and rows.", None))
#endif // QT_CONFIG(tooltip)
        self.compact_button.setText(QCoreApplication.translate("MainWindow", u"Compact", None))
        self.label_6.setText(QCoreApplication.translate("MainWindow", u"Group function:", None))
#if QT_CONFIG(tooltip)
        self.group_fn_combo_box.setToolTip(QCoreApplication.translate("MainWindow", u"Group/aggregate data that ends up in the same cell in pivot tables.", None))
#endif // QT_CONFIG(tooltip)
    # retranslateUi




######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains utilities to manage export manifest files."""
from itertools import dropwhile
import json
from pathlib import Path
from spine_engine.project_item.project_item_resource import file_resource_in_pack, transient_file_resource
from .specification import OutputFormat
from .utils import EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX


def exported_files_as_resources(item_name, exported_files, data_dir, output_channels, output_format):
    """Collects exported files from 'export manifests'.

    Args:
        item_name (str): item's name
        exported_files (dict, optional): item's exported files cache
        data_dir (str): item's data directory
        output_channels (Iterable of OutputChannel): item's output channels
        output_format (OutputFormat, optional): output format

    Returns:
        tuple: output resources and updated exported files cache
    """
    manifests = _collect_execution_manifests(data_dir)
    if output_format == OutputFormat.SQL:
        output_channels = tuple(c for c in output_channels if c.out_url is None)
    if manifests is not None:
        out_labels = {c.out_label for c in output_channels}
        manifests = {
            label: files for label, files in _collect_execution_manifests(data_dir).items() if label in out_labels
        }
        exported_files = {label: [str(Path(data_dir, f)) for f in files] for label, files in manifests.items()}
    resources = []
    if exported_files is not None:
        for channel in output_channels:
            if channel.out_label:
                files = {f for f in exported_files.get(channel.out_label, []) if Path(f).exists()}
                if files:
                    resources += [file_resource_in_pack(item_name, channel.out_label, f) for f in files]
                else:
                    resources.append(transient_file_resource(item_name, channel.out_label))
    else:
        for channel in output_channels:
            if channel.out_label:
                resources.append(transient_file_resource(item_name, channel.out_label))
    return resources, exported_files


def _collect_execution_manifests(data_dir):
    """Collects output file names from export manifest files written by exporter's executable item.

    Args:
        data_dir (str): item's data directory

    Returns:
        dict: mapping from output label to list of file paths, or None if no manifest files were found
    """
    manifests = None
    for path in Path(data_dir).iterdir():
        if is_manifest_file(path):
            with open(path) as manifest_file:
                try:
                    manifest = json.load(manifest_file)
                except json.decoder.JSONDecodeError:
                    continue
            for out_file_name, paths in manifest.items():
                relative_paths = []
                for file_path in paths:
                    p = Path(file_path)
                    if p.is_absolute():
                        # Legacy manifests had absolute paths
                        try:
                            relative_paths.append(p.relative_to(data_dir))
                        except ValueError:
                            # Project may have been moved to another directory (or system)
                            # so data_dir differs from manifest file content.
                            # Try resolving the relative path manually.
                            parts = tuple(dropwhile(lambda part: part != "output", p.parts))
                            relative_paths.append(str(Path(*parts)))
                    else:
                        relative_paths.append(file_path)
                if manifests is None:
                    manifests = {}
                manifests.setdefault(out_file_name, []).extend(paths)
    return manifests


def is_manifest_file(path):
    """Tests if given path is a manifest file.

    Args:
        path (Path): path to test

    Returns:
        bool: True if path looks like a manifest file, False otherwise
    """
    return path.is_file() and path.name.startswith(EXPORTER_EXECUTION_MANIFEST_FILE_PREFIX) and path.suffix == ".json"



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Data transformer's executable item as well as support utilities."""
from spine_engine.project_item.executable_item_base import ExecutableItemBase
from spine_engine.spine_engine import ItemExecutionFinishState
from .filter_config_path import filter_config_path
from .item_info import ItemInfo
from .output_resources import scan_for_resources


class ExecutableItem(ExecutableItemBase):
    def __init__(self, name, specification, project_dir, logger):
        """
        Args:
            name (str): item's name
            specification (DataTransformerSpecification, optional): item's specification
            project_dir (str): absolute path to project directory
            logger (LoggerInterface): a logger
        """
        super().__init__(name, project_dir, logger)
        self._db_resources = []
        self._specification = specification
        self._filter_config_path = filter_config_path(self._data_dir)

    @staticmethod
    def item_type():
        """Returns the data store executable's type identifier string."""
        return ItemInfo.item_type()

    def _output_resources_forward(self):
        """See base class."""
        return scan_for_resources(self, self._specification, self._db_resources, self._filter_config_path)

    def ready_to_execute(self, settings):
        if self._specification is None:
            return True
        messages = self._specification.settings.report_inconsistencies()
        if messages:
            for message in messages:
                self._logger.msg_error.emit(message)
            return False
        return True

    def execute(self, forward_resources, backward_resources, lock):
        """See base class."""
        if not super().execute(forward_resources, backward_resources, lock):
            return ItemExecutionFinishState.FAILURE
        self._db_resources = [r for r in forward_resources if r.type_ == "database"]
        return ItemExecutionFinishState.SUCCESS

    def exclude_execution(self, forward_resources, backward_resources, lock):
        """See base class."""
        self.execute(forward_resources, backward_resources, lock)

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        specification_name = item_dict["specification"]
        specification = ExecutableItemBase._get_specification(
            name, ItemInfo.item_type(), specification_name, specifications, logger
        )
        return cls(name, specification, project_dir, logger)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for ViewExecutable."""
from multiprocessing import Lock
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from spine_engine import ExecutionDirection
from spine_items.view.executable_item import ExecutableItem


class TestViewExecutable(unittest.TestCase):
    def setUp(self):
        self._temp_dir = TemporaryDirectory()

    def tearDown(self):
        self._temp_dir.cleanup()

    def test_item_type(self):
        self.assertEqual(ExecutableItem.item_type(), "View")

    def test_from_dict(self):
        logger = mock.MagicMock()
        item_dict = {"type": "View", "description": "", "x": 0, "y": 0}
        item = ExecutableItem.from_dict(item_dict, "Viewer", self._temp_dir.name, None, {}, logger)
        self.assertIsInstance(item, ExecutableItem)
        self.assertEqual("View", item.item_type())

    def test_stop_execution(self):
        executable = ExecutableItem(name="Viewer", project_dir=self._temp_dir.name, logger=mock.MagicMock())
        with mock.patch(
            "spine_engine.project_item.executable_item_base.ExecutableItemBase.stop_execution"
        ) as mock_stop_execution:
            executable.stop_execution()
            mock_stop_execution.assert_called_once()

    def test_execute_backward(self):
        executable = ExecutableItem("name", self._temp_dir.name, mock.MagicMock())
        self.assertTrue(executable.execute([], [], Lock()))

    def test_execute_forward(self):
        executable = ExecutableItem("name", self._temp_dir.name, mock.MagicMock())
        self.assertTrue(executable.execute([], [], Lock()))

    def test_output_resources_backward(self):
        executable = ExecutableItem("name", self._temp_dir.name, mock.MagicMock())
        self.assertEqual(executable.output_resources(ExecutionDirection.BACKWARD), [])

    def test_output_resources_forward(self):
        executable = ExecutableItem("name", self._temp_dir.name, mock.MagicMock())
        self.assertEqual(executable.output_resources(ExecutionDirection.FORWARD), [])


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains utilities to scan for Data Store's output resources."""
from spine_engine.project_item.project_item_resource import database_resource
from spine_items.utils import database_label


def scan_for_resources(provider, url):
    """
    Creates db resources based on DS url.

    Args:
        provider (ProjectItem or ExecutableItem): resource provider item
        url (URL): sqlalchemy url

    Returns:
        list of ProjectItemResource: output resources
    """
    if not url:
        return []
    return [database_resource(provider.name, str(url), label=database_label(provider.name), filterable=True)]



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Importer's executable item as well as support utilities."""
from contextlib import ExitStack
import os
from spine_engine.project_item.executable_item_base import ExecutableItemBase
from spine_engine.project_item.project_item_resource import get_labelled_source_resources
from spine_engine.spine_engine import ItemExecutionFinishState
from spine_engine.utils.returning_process import ReturningProcess
from spinedb_api.exception import ReaderError
from spinedb_api.spine_io.gdx_utils import find_gams_directory
from spinedb_api.spine_io.importers.csv_reader import CSVReader
from spinedb_api.spine_io.importers.datapackage_reader import DatapackageReader
from spinedb_api.spine_io.importers.excel_reader import ExcelReader
from spinedb_api.spine_io.importers.gdx_reader import GDXReader
from spinedb_api.spine_io.importers.json_reader import JSONReader
from spinedb_api.spine_io.importers.sqlalchemy_reader import SQLAlchemyReader
from ..db_writer_executable_item_base import DBWriterExecutableItemBase
from .do_work import do_work
from .item_info import ItemInfo

READER_NAME_TO_CLASS = {
    klass.__name__: klass
    for klass in (CSVReader, ExcelReader, GDXReader, JSONReader, DatapackageReader, SQLAlchemyReader)
}


class ExecutableItem(DBWriterExecutableItemBase):
    def __init__(self, name, mapping, selected_files, gams_path, cancel_on_error, on_conflict, project_dir, logger):
        """
        Args:
            name (str): Importer's name
            mapping (dict): import mapping
            selected_files (list): selected_files
            gams_path (str): path to system's GAMS executable or empty string for the default path
            cancel_on_error (bool): if True, revert changes on error and quit
            on_conflict (str): conflict resolution strategy for spinedb_api.import_data
            project_dir (str): absolute path to project directory
            logger (LoggerInterface): a logger
        """
        super().__init__(name, project_dir, logger)
        self._mapping = mapping
        self._selected_files = selected_files
        self._gams_path = gams_path
        self._cancel_on_error = cancel_on_error
        self._on_conflict = on_conflict
        self._process = None

    @staticmethod
    def item_type():
        """Returns ImporterExecutable's type identifier string."""
        return ItemInfo.item_type()

    def stop_execution(self):
        """Stops executing this Importer."""
        super().stop_execution()
        if self._process is not None:
            self._process.terminate()
            self._process = None

    def execute(self, forward_resources, backward_resources, lock):
        """See base class."""
        if not super().execute(forward_resources, backward_resources, lock):
            return ItemExecutionFinishState.FAILURE
        if not self._mapping:
            self._logger.msg_warning.emit(f"No mappings configured. Skipping.")
            return ItemExecutionFinishState.SKIPPED
        labelled_resources = get_labelled_source_resources(forward_resources)
        selected_resources = []
        for label in self._selected_files:
            selected_resources += labelled_resources.get(label, [])
        to_resources = [r for r in backward_resources if r.type_ == "database"]
        if not selected_resources or not to_resources:
            return ItemExecutionFinishState.SUCCESS
        source_type = self._mapping["source_type"]
        if source_type == "GDXReader":
            source_settings = {"gams_directory": self._gams_system_directory()}
        else:
            source_settings = None
        try:
            connector = READER_NAME_TO_CLASS[source_type](source_settings)
        except ReaderError as error:
            self._logger.msg_error.emit(f"Failed to create reader: {error}")
            return ItemExecutionFinishState.FAILURE
        with ExitStack() as stack:
            to_server_urls = [stack.enter_context(resource.open()) for resource in to_resources]
            self._process = ReturningProcess(
                target=do_work,
                args=(
                    self._mapping,
                    self._cancel_on_error,
                    self._on_conflict,
                    self._logs_dir,
                    selected_resources,
                    connector,
                    to_server_urls,
                    lock,
                    self._logger,
                ),
            )
            return_value = self._process.run_until_complete()
            self._process = None
            return ItemExecutionFinishState.SUCCESS if return_value[0] else ItemExecutionFinishState.FAILURE

    def _gams_system_directory(self):
        """Returns GAMS system path or None if GAMS default is to be used."""
        path = self._gams_path
        if not path:
            path = find_gams_directory()
        if path is not None and os.path.isfile(path):
            path = os.path.dirname(path)
        return path

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        specification_name = item_dict["specification"]
        specification = ExecutableItemBase._get_specification(
            name, ItemInfo.item_type(), specification_name, specifications, logger
        )
        mapping = specification.mapping if specification else {}
        file_selection = dict(item_dict.get("file_selection", []))
        selected_files = [filepath for filepath, selected in file_selection.items() if selected]
        gams_path = app_settings.value("appSettings/gamsPath", defaultValue=None)
        cancel_on_error = item_dict["cancel_on_error"]
        on_conflict = item_dict["on_conflict"]
        return cls(name, mapping, selected_files, gams_path, cancel_on_error, on_conflict, project_dir, logger)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains mapping filter editor delegate."""
from spine_items.ui import default_filter_editor
from spine_items.widgets import FilterEdit, FilterEditDelegateBase
from spinedb_api.export_mapping.export_mapping import ParameterValueTypeMapping
from ..mvcmodels.mapping_editor_table_model import MappingEditorTableModel
from ..ui import value_type_filter_editor


class FilterEditDelegate(FilterEditDelegateBase):
    """Edit delegate for Mapping table's filter column."""

    def createEditor(self, parent, option, index):
        mapping = index.data(MappingEditorTableModel.MAPPING_ITEM_ROLE)
        if isinstance(mapping, ParameterValueTypeMapping):
            return FilterEdit(value_type_filter_editor.Ui_Form(), parent)
        return FilterEdit(default_filter_editor.Ui_Form(), parent)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################
import unittest
from unittest import mock
from spine_items.data_store.commands import UpdateDSURLCommand


class TestUpdateDSURLCommand(unittest.TestCase):
    def test_merging_with_command_that_modifies_same_fields_works(self):
        mock_ds = mock.MagicMock()
        mock_ds.url.return_value = {"username": "c"}
        mock_project = mock.MagicMock()
        mock_project.get_item.return_value = mock_ds
        command1 = UpdateDSURLCommand("DS 1", False, mock_project, username="ca")
        mock_ds.url.return_value = {"username": "ca"}
        command2 = UpdateDSURLCommand("DS 1", False, mock_project, username="cat")
        self.assertTrue(command1.mergeWith(command2))
        command1.redo()
        mock_ds.do_update_url.assert_called_once_with(username="cat")

    def test_merging_is_rejected_when_commands_modify_different_fields(self):
        mock_ds = mock.MagicMock()
        mock_ds.url.return_value = {"username": "c", "host": ""}
        mock_project = mock.MagicMock()
        mock_project.get_item.return_value = mock_ds
        command1 = UpdateDSURLCommand("DS 1", False, mock_project, username="ca")
        mock_ds.url.return_value = {"username": "ca", "host": ""}
        command2 = UpdateDSURLCommand("DS 1", False, mock_project, host="q")
        self.assertFalse(command1.mergeWith(command2))


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for Importer project item."""
import collections
import os
from tempfile import TemporaryDirectory
import unittest
from unittest.mock import MagicMock, NonCallableMagicMock
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication, QMenu
from spine_engine.project_item.project_item_resource import file_resource, transient_file_resource
from spine_items.importer.importer import Importer
from spine_items.importer.importer_factory import ImporterFactory
from spine_items.importer.importer_specification import ImporterSpecification
from spine_items.importer.item_info import ItemInfo
from tests.mock_helpers import create_mock_project, create_mock_toolbox, mock_finish_project_item_construction


class TestImporter(unittest.TestCase):
    def setUp(self):
        """Set up."""
        self.toolbox = create_mock_toolbox()
        mock_spec_model = self.toolbox.specification_model = MagicMock()
        Specification = collections.namedtuple("Specification", "name mapping item_type")
        mock_spec_model.find_specification.side_effect = lambda x: Specification(
            name=x, mapping={}, item_type="Importer"
        )
        factory = ImporterFactory()
        item_dict = {
            "type": "Importer",
            "description": "Very best test importer",
            "specification": "importer_spec",
            "cancel_on_error": True,
            "file_selection": [["file 99", True], ["file 01", False]],
            "x": 0,
            "y": 0,
        }
        self._temp_dir = TemporaryDirectory()
        self.project = create_mock_project(self._temp_dir.name)
        self.toolbox.project.return_value = self.project
        self.importer = factory.make_item("I", item_dict, self.toolbox, self.project)
        self._properties_widget = mock_finish_project_item_construction(factory, self.importer, self.toolbox)

    def tearDown(self):
        self._temp_dir.cleanup()

    @classmethod
    def setUpClass(cls):
        if not QApplication.instance():
            QApplication()

    def test_item_type(self):
        self.assertEqual(Importer.item_type(), ItemInfo.item_type())

    def test_item_dict(self):
        """Tests Item dictionary creation."""
        specification = ImporterSpecification("import specification", {})
        self.importer.do_set_specification(specification)
        resources = [
            file_resource("source 1", "/path/file1.dat", "<source 1>/file1.dat"),
            file_resource("source 2", "/path/file2.dat", "<source 2>/file2.dat"),
        ]
        self.importer.upstream_resources_updated(resources)
        index = self.importer._file_model.index(1, 0)
        self.importer._file_model.setData(index, False, Qt.ItemDataRole.CheckStateRole)
        d = self.importer.item_dict()
        self.assertEqual(
            d,
            {
                "cancel_on_error": True,
                "description": "Very best test importer",
                "file_selection": [["<source 1>/file1.dat", True], ["<source 2>/file2.dat", False]],
                "on_conflict": "merge",
                "specification": "import specification",
                "type": "Importer",
                "x": 0.0,
                "y": 0.0,
            },
        )

    def test_notify_destination(self):
        self.importer.logger.msg = MagicMock()
        self.importer.logger.msg_warning = MagicMock()
        source_item = NonCallableMagicMock()
        source_item.name = "source name"
        source_item.item_type = MagicMock(return_value="Data Connection")
        self.importer.notify_destination(source_item)
        self.importer.logger.msg.emit.assert_called_with(
            "Link established. You can define mappings on data from <b>source name</b> using <b>I</b>."
        )
        source_item.item_type = MagicMock(return_value="Data Store")
        self.importer.notify_destination(source_item)
        self.importer.logger.msg.emit.assert_called_with("Link established")
        source_item.item_type = MagicMock(return_value="Data Transformer")
        self.importer.notify_destination(source_item)
        self.importer.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a "
            "<b>Data Transformer</b> and a <b>Importer</b> has not been implemented yet."
        )
        source_item.item_type = MagicMock(return_value="Exporter")
        self.importer.notify_destination(source_item)
        self.importer.logger.msg.emit.assert_called_with(
            "Link established. You can define mappings on data from <b>source name</b> using <b>I</b>."
        )
        source_item.item_type = MagicMock(return_value="Tool")
        self.importer.notify_destination(source_item)
        self.importer.logger.msg.emit.assert_called_with(
            "Link established. You can define mappings on data from <b>source name</b> using <b>I</b>."
        )
        source_item.item_type = MagicMock(return_value="View")
        self.importer.notify_destination(source_item)
        self.importer.logger.msg_warning.emit.assert_called_with(
            "Link established. Interaction between a <b>View</b> and a <b>Importer</b> has not been implemented yet."
        )

    def test_rename(self):
        """Tests renaming an Importer."""
        with unittest.mock.patch("spine_items.importer.importer.ItemSpecificationMenu") as mock_spec_menu:
            mock_spec_menu.return_value = QMenu()
            self.importer.activate()
        expected_name = "ABC"
        expected_short_name = "abc"
        self.importer.rename(expected_name, "")
        # Check name
        self.assertEqual(expected_name, self.importer.name)  # item name
        self.assertEqual(expected_name, self.importer.get_icon().name())  # name item on Design View
        # Check data_dir
        expected_data_dir = os.path.join(self.project.items_dir, expected_short_name)
        self.assertEqual(expected_data_dir, self.importer.data_dir)  # Check data dir

    def test_handle_dag_changed(self):
        """Tests that upstream resource files are listed in the Importer view."""
        with unittest.mock.patch("spine_items.importer.importer.ItemSpecificationMenu") as mock_spec_menu:
            mock_spec_menu.return_value = QMenu()
            self.importer.activate()
        expected_file_list = ["file1", "file2"]
        resources = [transient_file_resource("provider", label=file_name) for file_name in expected_file_list]
        self.importer.upstream_resources_updated(resources)
        model = self.importer._properties_ui.treeView_files.model()
        file_list = [model.index(row, 0).data(Qt.ItemDataRole.DisplayRole) for row in range(model.rowCount())]
        self.assertEqual(sorted(file_list), sorted(expected_file_list))
        checked = [model.index(row, 0).data(Qt.ItemDataRole.CheckStateRole).value for row in range(model.rowCount())]
        selected = [check == Qt.CheckState.Checked.value for check in checked]
        self.assertTrue(all(selected))

    def test_handle_dag_changed_updates_previous_list_items(self):
        with unittest.mock.patch("spine_items.importer.importer.ItemSpecificationMenu") as mock_spec_menu:
            mock_spec_menu.return_value = QMenu()
            self.importer.activate()
        resources = [transient_file_resource("provider", label=name) for name in ["file1", "file2"]]
        # Add initial files
        self.importer.upstream_resources_updated(resources)
        model = self.importer._properties_ui.treeView_files.model()
        for row in range(2):
            index = model.index(row, 0)
            model.setData(index, Qt.CheckState.Unchecked, Qt.ItemDataRole.CheckStateRole)
        # Update with one existing, one new file
        resources = [transient_file_resource("provider", label=name) for name in ["file2", "file3"]]
        self.importer.upstream_resources_updated(resources)
        file_list = [model.index(row, 0).data(Qt.ItemDataRole.DisplayRole) for row in range(model.rowCount())]
        self.assertEqual(file_list, ["file2", "file3"])
        checked = [model.index(row, 0).data(Qt.ItemDataRole.CheckStateRole).value for row in range(model.rowCount())]
        selected = [check == Qt.CheckState.Checked.value for check in checked]
        self.assertEqual(selected, [False, True])


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Provides an optional widget for Tool Specification Editor when editing Julia, Python, or Executable Tool Specs."""
import sys
from PySide6.QtCore import Slot
from PySide6.QtGui import QIcon, QStandardItem, QStandardItemModel, Qt
from PySide6.QtWidgets import QApplication, QWidget
from spine_engine.utils.helpers import resolve_default_julia_executable, resolve_python_interpreter
from spinetoolbox.helpers import file_is_valid, select_julia_executable, select_julia_project, select_python_interpreter
from spinetoolbox.kernel_fetcher import KernelFetcher
from spinetoolbox.widgets.notification import Notification


class OptionalWidget(QWidget):
    def __init__(self, specification_editor, toolbox):
        """
        Args:
            specification_editor (ToolSpecificationEditorWindow): Tool spec editor window
            toolbox (ToolboxUI): Toolbox main window
        """
        super().__init__()
        self._toolbox = toolbox
        self._specification_editor = specification_editor

    def init_widget(self, specification):
        raise NotImplementedError()

    def add_execution_settings(self, tool_spec_type):
        """Collects execution settings based on optional widget state into a dictionary, which is returned.

        Args:
            tool_spec_type (str): Tool spec type
        """
        raise NotImplementedError()

    def default_execution_settings(self):
        """Returns default execution settings dictionary."""
        raise NotImplementedError()

    def get_widgets_in_tab_order(self):
        """Returns widget in tab order.

        Returns:
            Sequence of QWidget: widgets
        """
        raise NotImplementedError()


class SharedToolSpecOptionalWidget(OptionalWidget):
    """Superclass for Python and Julia Tool Spec optional widgets."""

    def __init__(self, specification_editor, toolbox, Ui_Form, fetch_mode):
        """
        Args:
            specification_editor (ToolSpecificationEditorWindow): Tool spec editor window
            toolbox (ToolboxUI): Toolbox main window
            Ui_Form (Form): Optional widget UI form
            fetch_mode (int): Kernel fetch mode (see KernelFetcher class)
        """
        super().__init__(specification_editor, toolbox)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.fetch_mode = fetch_mode
        self.conda_path = self._toolbox.qsettings().value("appSettings/condaPath", defaultValue="")
        self.kernel_spec_model = QStandardItemModel(self)
        self.ui.comboBox_kernel_specs.setModel(self.kernel_spec_model)
        self._kernel_spec_editor = None
        self._kernel_spec_model_initialized = False
        self._saved_kernel = None
        self._selected_kernel = None
        self.kernel_fetcher = None

    def connect_signals(self):
        """Connects signals."""
        self.ui.toolButton_refresh_kernel_specs.clicked.connect(self.start_kernel_fetcher)
        self.ui.comboBox_kernel_specs.activated.connect(self._specification_editor.push_change_kernel_spec_command)
        self.ui.radioButton_jupyter_console.toggled.connect(self._specification_editor.push_set_jupyter_console_mode)
        self.ui.lineEdit_executable.textEdited.connect(self._specification_editor.push_change_executable)
        self.ui.lineEdit_executable.editingFinished.connect(self._specification_editor.finish_changing_executable)
        qApp.aboutToQuit.connect(self.stop_fetching_kernels)  # pylint: disable=undefined-variable

    def init_widget(self, specification):
        """Initializes UI elements based on specification

        Args:
            specification (ToolSpecification): Specification to load
        """
        use_jupyter_console = specification.execution_settings["use_jupyter_console"]
        self.ui.radioButton_jupyter_console.blockSignals(True)
        self.ui.radioButton_basic_console.blockSignals(True)
        if use_jupyter_console:
            self.ui.radioButton_jupyter_console.setChecked(True)
        else:
            self.ui.radioButton_basic_console.setChecked(True)
        self.ui.radioButton_jupyter_console.blockSignals(False)
        self.ui.radioButton_basic_console.blockSignals(False)
        self._saved_kernel = specification.execution_settings["kernel_spec_name"]
        self.set_executable(specification.execution_settings["executable"])
        self.set_ui_for_jupyter_console(use_jupyter_console)

    def set_ui_for_jupyter_console(self, use_jupyter_console):
        """Enables or disables some UI elements in the optional widget according to a checkBox state.

        Args:
            use_jupyter_console (bool): True when Jupyter Console checkBox is checked, false otherwise
        """
        self.ui.lineEdit_executable.setEnabled(not use_jupyter_console)  # Disable for jupyter console
        self.ui.comboBox_kernel_specs.setEnabled(use_jupyter_console)  # Enable for jupyter console
        self.ui.toolButton_refresh_kernel_specs.setEnabled(use_jupyter_console)  # Enable for jupyter console
        if use_jupyter_console and not self._kernel_spec_model_initialized:
            self.start_kernel_fetcher(restore_saved_kernel=True)

    def add_execution_settings(self, tool_spec_type):
        """See base class."""
        idx = self.ui.comboBox_kernel_specs.currentIndex()
        if idx < 1:
            d = {"kernel_spec_name": "", "env": ""}
        else:
            item = self.ui.comboBox_kernel_specs.model().item(idx)
            k_spec_data = item.data()
            d = k_spec_data
        d["use_jupyter_console"] = self.ui.radioButton_jupyter_console.isChecked()
        p = self.get_executable()
        self.validate_executable(p, tool_spec_type)  # Raises NameError if Python or Julia path is not valid
        d["executable"] = p
        return d

    def validate_executable(self, p, tool_spec_type):
        """Check that given Python or Julia path is a file, it exists, and the
        file name starts with either 'python' or 'julia'.

        Args:
            p (str): Abs. path to check
            tool_spec_type (str): 'python' or 'julia'

        Raises:
            NameError: If the python path in the line edit is not valid
        """
        if not file_is_valid(
            self._specification_editor,
            p,
            f"Invalid {tool_spec_type.capitalize()} Interpreter",
            extra_check=tool_spec_type,
        ):
            raise NameError

    def set_executable(self, p):
        """Sets given path p to either Python or Julia line edit."""
        if p == self.get_executable():
            return
        self.ui.lineEdit_executable.setText(p)

    def get_executable(self):
        """Returns either Python or Julia executable path."""
        return self.ui.lineEdit_executable.text().strip()

    def find_index_by_data(self, string):
        """Searches the kernel spec model for the first item whose data matches the given string.
        Returns the items row number or -1 if not found."""
        if string == "":
            return 0
        for row in range(1, self.kernel_spec_model.rowCount()):  # Start from index 1
            index = self.kernel_spec_model.index(row, 0)
            item = self.kernel_spec_model.itemFromIndex(index)
            d = item.data()
            if d["kernel_spec_name"] == string:
                return row
        return -1

    @Slot()
    def start_kernel_fetcher(self, restore_saved_kernel=False):
        """Initializes kernel spec model and starts KernelFetcher.

        Args:
            restore_saved_kernel (bool): True restores a saved kernel after the model has been refreshed,
            False restores the selected kernel after the model has been refreshed.
        """
        if self.kernel_fetcher is not None and self.kernel_fetcher.isRunning():
            return
        QApplication.setOverrideCursor(Qt.CursorShape.BusyCursor)
        self.initialize_kernel_spec_model()
        self.kernel_fetcher = KernelFetcher(self.conda_path, fetch_mode=self.fetch_mode)
        self.kernel_fetcher.kernel_found.connect(self.add_kernel)
        if restore_saved_kernel:
            self.kernel_fetcher.finished.connect(self._restore_saved_kernel)
        else:
            self.kernel_fetcher.finished.connect(self._restore_selected_kernel)
        self.kernel_fetcher.finished.connect(self._toolbox.restore_override_cursor)
        self.kernel_fetcher.start()

    @Slot()
    def stop_fetching_kernels(self):
        """Terminates the kernel fetcher thread."""
        if self.kernel_fetcher is not None:
            self.kernel_fetcher.stop_fetcher.emit()

    @Slot(str, str, bool, QIcon, dict)
    def add_kernel(self, kernel_name, resource_dir, conda, icon, deats):
        """Adds a kernel entry as an item to kernel spec model."""
        if not conda:
            spec_data = {"kernel_spec_name": kernel_name, "env": ""}
            item = QStandardItem(kernel_name)
        else:
            spec_data = {"kernel_spec_name": kernel_name, "env": "conda"}
            item = QStandardItem(kernel_name + " [Conda]")
        item.setIcon(icon)
        item.setData(spec_data)
        self.kernel_spec_model.appendRow(item)

    def initialize_kernel_spec_model(self):
        """Initializes kernel spec model for fetching kernels. Memorizes currently
        selected kernel in order to select it again after the model has been
        refreshed."""
        item = self.kernel_spec_model.item(self.ui.comboBox_kernel_specs.currentIndex())
        if not item or not item.data():
            self._selected_kernel = None
        else:
            self._selected_kernel = item.data()["kernel_spec_name"]  # Remember the selected kernel spec
        self.kernel_spec_model.clear()
        first_item = QStandardItem("Select kernel spec...")
        self.kernel_spec_model.appendRow(first_item)

    @Slot()
    def _restore_selected_kernel(self):
        """Sets the previously selected kernel spec as the current item after the model has been refreshed."""
        self._kernel_spec_model_initialized = True
        if self._selected_kernel is not None:
            row = self.find_index_by_data(self._selected_kernel)
            if row == -1:
                # The kernel spec may have been removed
                self._specification_editor.push_change_kernel_spec_command(0)
                return
            self.ui.comboBox_kernel_specs.setCurrentIndex(row)

    @Slot()
    def _restore_saved_kernel(self):
        """Sets index of the kernel spec combobox to show the item that was saved with the Tool Specification.
        Make sure this is called after kernel spec model has been populated."""
        self._kernel_spec_model_initialized = True
        if not self._saved_kernel:
            self.ui.comboBox_kernel_specs.setCurrentIndex(0)  # Set 'Select kernel spec...'
        else:
            row = self.find_index_by_data(self._saved_kernel)
            if row == -1:
                notification = Notification(
                    self._specification_editor,
                    f"This Tool spec has kernel spec '{self._saved_kernel}' " f"saved but it could not be found.",
                )
                notification.show()
                row += 1  # Set 'Select kernel spec...'
            self.ui.comboBox_kernel_specs.setCurrentIndex(row)


class PythonToolSpecOptionalWidget(SharedToolSpecOptionalWidget):
    def __init__(self, specification_editor, toolbox):
        """
        Args:
            specification_editor (ToolSpecificationEditorWindow): Tool spec editor window
            toolbox (ToolboxUI): Toolbox main window
        """
        from ..ui.python_kernel_spec_options import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(specification_editor, toolbox, Ui_Form, 2)
        # Initialize UI elements with defaults
        use_jupyter_console = bool(
            int(self._toolbox.qsettings().value("appSettings/usePythonKernel", defaultValue="0"))
        )
        if use_jupyter_console:
            self.ui.radioButton_jupyter_console.setChecked(True)
            # Get the default kernel spec from Settings->Tools for new Jupyter Console Tool Specs
            self._saved_kernel = self._toolbox.qsettings().value("appSettings/pythonKernel", defaultValue="")
        else:
            self.ui.radioButton_basic_console.setChecked(True)
        default_python_path = resolve_python_interpreter(self._toolbox.qsettings())
        self.ui.lineEdit_executable.setPlaceholderText(default_python_path)
        self.set_ui_for_jupyter_console(use_jupyter_console)
        self.connect_signals()

    def connect_signals(self):
        """Connects signals to slots."""
        super().connect_signals()
        self.ui.toolButton_browse_python.clicked.connect(self.browse_python_button_clicked)

    def default_execution_settings(self):
        """See base class."""
        use_jupyter_cons = bool(int(self._toolbox.qsettings().value("appSettings/usePythonKernel", defaultValue="0")))
        k_name = self._toolbox.qsettings().value("appSettings/pythonKernel", defaultValue="")
        env = ""
        if use_jupyter_cons:
            # Check if the kernel is a Conda kernel by matching the name with the one that is in kernel_spec_model
            # Find k_name in kernel_spec_model and check it's data
            row = self.find_index_by_data(k_name)
            if row == -1:
                pass  # kernel not found
            else:
                index = self.kernel_spec_model.index(row, 0)
                item_data = self.kernel_spec_model.itemFromIndex(index).data()
                env = item_data["env"]
        d = {}
        d["kernel_spec_name"] = k_name
        d["env"] = env
        d["use_jupyter_console"] = use_jupyter_cons
        d["executable"] = self._toolbox.qsettings().value("appSettings/pythonPath", defaultValue="")
        return d

    @Slot(bool)
    def browse_python_button_clicked(self, _=False):
        """Calls static method that shows a file browser for selecting a Python interpreter."""
        select_python_interpreter(self, self.ui.lineEdit_executable)
        self._specification_editor.push_change_executable(self.get_executable())
        self._specification_editor.finish_changing_executable()

    def set_ui_for_jupyter_console(self, use_jupyter_console):
        """Enables or disables some UI elements in the optional widget according to a checkBox state.

        Args:
            use_jupyter_console (bool): True when Jupyter Console checkBox is checked, false otherwise
        """
        self.ui.toolButton_browse_python.setEnabled(not use_jupyter_console)  # Disable for jupyter console
        super().set_ui_for_jupyter_console(use_jupyter_console)

    def get_widgets_in_tab_order(self):
        """See base class."""
        return (
            self.ui.radioButton_basic_console,
            self.ui.radioButton_jupyter_console,
            self.ui.lineEdit_executable,
            self.ui.toolButton_browse_python,
            self.ui.comboBox_kernel_specs,
            self.ui.toolButton_refresh_kernel_specs,
        )


class JuliaToolSpecOptionalWidget(SharedToolSpecOptionalWidget):
    def __init__(self, specification_editor, toolbox):
        """
        Args:
            specification_editor (ToolSpecificationEditorWindow): Tool spec editor window
            toolbox (ToolboxUI): Toolbox main window
        """
        from ..ui.julia_kernel_spec_options import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(specification_editor, toolbox, Ui_Form, 4)
        # Initialize UI elements with defaults
        use_jupyter_console = bool(int(self._toolbox.qsettings().value("appSettings/useJuliaKernel", defaultValue="0")))
        if use_jupyter_console:
            self.ui.radioButton_jupyter_console.setChecked(True)
            # Get the default kernel spec from Settings->Tools for new Jupyter Console Tool Specs
            self._saved_kernel = self._toolbox.qsettings().value("appSettings/juliaKernel", defaultValue="")
        else:
            self.ui.radioButton_basic_console.setChecked(True)
        default_julia_path = self._toolbox.qsettings().value("appSettings/juliaPath", defaultValue="")
        default_julia_project = self._toolbox.qsettings().value("appSettings/juliaProjectPath", defaultValue="")
        if not default_julia_path:
            default_julia_path = resolve_default_julia_executable()
        self.ui.lineEdit_executable.setPlaceholderText(default_julia_path)
        self.ui.lineEdit_julia_project.setPlaceholderText(default_julia_project)
        self.set_ui_for_jupyter_console(use_jupyter_console)
        self.connect_signals()

    def connect_signals(self):
        """Connects signals."""
        super().connect_signals()
        self.ui.toolButton_browse_julia.clicked.connect(self.browse_julia_button_clicked)
        self.ui.toolButton_browse_julia_project.clicked.connect(self.browse_julia_project_button_clicked)
        self.ui.lineEdit_julia_project.editingFinished.connect(self._specification_editor.push_change_project)

    def init_widget(self, specification):
        """Initializes UI elements based on specification

        Args:
            specification (ToolSpecification): Specification to load
        """
        project = specification.execution_settings["project"]
        self.ui.lineEdit_julia_project.setText(project)
        super().init_widget(specification)

    def add_execution_settings(self, tool_spec_type):
        """See base class."""
        d = super().add_execution_settings(tool_spec_type)
        d["project"] = self.get_julia_project()
        return d

    def default_execution_settings(self):
        """See base class."""
        d = {}
        use_jupyter_console = bool(int(self._toolbox.qsettings().value("appSettings/useJuliaKernel", defaultValue="0")))
        d["kernel_spec_name"] = self._toolbox.qsettings().value("appSettings/juliaKernel", defaultValue="")
        d["env"] = ""
        d["use_jupyter_console"] = use_jupyter_console
        d["executable"] = self._toolbox.qsettings().value("appSettings/juliaPath", defaultValue="")
        d["project"] = self._toolbox.qsettings().value("appSettings/juliaProjectPath", defaultValue="")
        return d

    @Slot(bool)
    def browse_julia_button_clicked(self, _=False):
        """Calls static method that shows a file browser for selecting a Julia executable."""
        select_julia_executable(self, self.ui.lineEdit_executable)
        self._specification_editor.push_change_executable(self.get_executable())

    @Slot(bool)
    def browse_julia_project_button_clicked(self, _=False):
        """Calls static method that shows a file browser for selecting a Julia project."""
        select_julia_project(self, self.ui.lineEdit_julia_project)
        self._specification_editor.push_change_project()

    def set_ui_for_jupyter_console(self, use_jupyter_console):
        """Enables or disables some UI elements in the optional widget according to a checkBox state.

        Args:
            use_jupyter_console (bool): True when Jupyter Console checkBox is checked, false otherwise
        """

        self.ui.lineEdit_julia_project.setEnabled(not use_jupyter_console)  # Disable for jupyter console
        self.ui.toolButton_browse_julia.setEnabled(not use_jupyter_console)  # Disable for jupyter console
        self.ui.toolButton_browse_julia_project.setEnabled(not use_jupyter_console)  # Disable for jupyter console
        super().set_ui_for_jupyter_console(use_jupyter_console)

    def set_julia_project(self, p):
        """Sets given Julia project to Julia project line edit."""
        self.ui.lineEdit_julia_project.setText(p)

    def get_julia_project(self):
        """Returns Julia project from line edit."""
        return self.ui.lineEdit_julia_project.text().strip()

    def get_widgets_in_tab_order(self):
        """See base class."""
        return (
            self.ui.radioButton_basic_console,
            self.ui.radioButton_jupyter_console,
            self.ui.lineEdit_executable,
            self.ui.toolButton_browse_julia,
            self.ui.lineEdit_julia_project,
            self.ui.toolButton_browse_julia_project,
            self.ui.comboBox_kernel_specs,
            self.ui.toolButton_refresh_kernel_specs,
        )


class ExecutableToolSpecOptionalWidget(OptionalWidget):
    def __init__(self, specification_editor, toolbox):
        """
        Args:
            specification_editor (ToolSpecificationEditorWindow): tool specification editor window
            toolbox (ToolboxUI): Toolbox main window
        """
        from ..ui.executable_cmd_exec_options import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(specification_editor, toolbox)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.shells = ["No shell", "cmd.exe", "powershell.exe", "bash"]
        self.ui.comboBox_shell.addItems(self.shells)
        # Disable shell selections that are not compatible with the os running toolbox
        msg = f"<p>Selection not available for platform {sys.platform}</p>"
        if sys.platform == "win32":
            # No bash for windows
            self.ui.comboBox_shell.model().item(3).setEnabled(False)
            self.ui.comboBox_shell.model().item(3).setToolTip(msg)
        else:
            # For other systems no cmd or powershell
            self.ui.comboBox_shell.model().item(1).setEnabled(False)
            self.ui.comboBox_shell.model().item(1).setToolTip(msg)
            self.ui.comboBox_shell.model().item(2).setEnabled(False)
            self.ui.comboBox_shell.model().item(2).setToolTip(msg)
        self.connect_signals()

    def connect_signals(self):
        self.ui.lineEdit_command.textEdited.connect(self._specification_editor.push_change_executable_command)
        self.ui.lineEdit_command.editingFinished.connect(self._specification_editor.finish_changing_executable)
        self.ui.comboBox_shell.activated.connect(self._specification_editor.push_change_shell_command)

    def init_widget(self, specification):
        """Initializes UI elements based on specification."""
        self.ui.lineEdit_command.setText(specification.execution_settings["cmd"])
        shell = specification.execution_settings["shell"]
        ind = next(iter(k for k, t in enumerate(self.shells) if t.lower() == shell), 0)
        self.ui.comboBox_shell.setCurrentIndex(ind)

    def add_execution_settings(self, tool_spec_type):
        """See base class."""
        return {"cmd": self.ui.lineEdit_command.text(), "shell": self.get_current_shell()}

    def get_current_shell(self):
        """Returns the selected shell in the shell combo box."""
        ind = self.ui.comboBox_shell.currentIndex()
        if ind < 1:
            return ""
        return self.ui.comboBox_shell.currentText()

    def set_command_and_shell_edit_disabled_state(self, enabled):
        """Sets the enabled state for the Command -text editor and the Shell -combobox"""
        self.ui.comboBox_shell.setDisabled(enabled)
        self.ui.lineEdit_command.setDisabled(enabled)

    def get_widgets_in_tab_order(self):
        """See base class."""
        return (
            self.ui.lineEdit_command,
            self.ui.comboBox_shell,
        )



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains a generic File list model and an Item for that model."""
from collections import namedtuple
from PySide6.QtCore import QModelIndex, Qt, Signal
from spine_engine.project_item.project_item_resource import extract_packs
from spinetoolbox.mvcmodels.file_list_models import (
    CommandLineArgItem,
    CommandLineArgsModel,
    FileListModel,
    NewCommandLineArgItem,
)


class CheckableFileListModel(FileListModel):
    """A model for checkable files to be shown in a file tree view."""

    FileItem = namedtuple("FileItem", ["resource", "checked"])
    PackItem = namedtuple("PackItem", ["label", "resources", "checked"])

    checked_state_changed = Signal(QModelIndex, bool)
    """Emitted when an item's checked state changes."""

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.CheckStateRole:
            if index.internalPointer() is None:
                row = index.row()
                if row < len(self._single_resources):
                    checked = self._single_resources[row].checked
                else:
                    checked = self._pack_resources[row - len(self._single_resources)].checked
                return Qt.CheckState.Checked if checked else Qt.CheckState.Unchecked
            return None
        return super().data(index, role)

    def checked_data(self, index):
        """Returns checked status and label for given top-level index.

        Args:
            index (QModelIndex): top-level index

        Returns:
            tuple: item label and checked flag
        """
        row = index.row()
        if row < len(self._single_resources):
            return self._single_resources[row].resource.label, self._single_resources[row].checked
        pack = self._pack_resources[row - len(self._single_resources)]
        return pack.label, pack.checked

    def is_checked(self, index):
        """Returns True if the item at given index is checked.

        Args:
            index (QModelIndex): index

        Returns:
            bool: True if the item is checked, False otherwise
        """
        pack_label = index.internalPointer()
        if pack_label is None:
            row = index.row()
            if row < len(self._single_resources):
                return self._single_resources[row].checked
            return self._pack_resources[row - len(self._single_resources)].checked
        return self._pack_resources[self._pack_index(pack_label)].checked

    def flags(self, index):
        flags = super().flags(index)
        if index.internalPointer() is None:
            return flags | Qt.ItemFlag.ItemIsUserCheckable
        return flags

    def update(self, resources):
        """Updates the model according to given list of resources.

        Args:
            resources (Iterable of ProjectItemResource): resources
        """
        self.beginResetModel()
        unchecked_singles = {item.resource.label for item in self._single_resources if not item.checked}
        unchecked_packs = {item.label for item in self._pack_resources if not item.checked}
        unchecked = unchecked_singles | unchecked_packs
        single_resources, pack_resources = extract_packs(resources)
        new_singles = [self.FileItem(r, r.label not in unchecked) for r in single_resources]
        new_packs = [
            self.PackItem(label, [r for r in r_list if r.hasfilepath], label not in unchecked)
            for label, r_list in pack_resources.items()
        ]
        self._single_resources = new_singles
        self._pack_resources = new_packs
        self.endResetModel()

    def replace(self, old_resource, new_resource):
        """Replaces existing data in the model.

        Note: we don't currently update file pack labels here.

        Args:
            old_resource (ProjectItemResource): resource to replace
            new_resource (ProjectItemResource): new resource
        """
        if old_resource.type_ == "database":
            return
        for row, item in enumerate(self._single_resources):
            if item.resource.provider_name == old_resource.provider_name and item.resource.label == old_resource.label:
                self._single_resources[row] = item._replace(resource=new_resource)
                index = self.index(row, 0)
                self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.ToolTipRole])
                return
        for pack_row, pack in enumerate(self._pack_resources):
            for row, resource in enumerate(pack.resources):
                if resource.provider_name == old_resource.provider_name and resource.label == old_resource.label:
                    if new_resource.label == resource.label:
                        pack.resources[row] = new_resource
                        pack_index = self.index(pack_row, 0)
                        index = self.index(row, 0, pack_index)
                        self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.ToolTipRole])
                        return
                    single_resources = [item.resource for item in self._single_resources]
                    new_pack_resources = [
                        r
                        for existing_pack in self._pack_resources
                        for r in existing_pack.resources
                        if r is not resource
                    ]
                    new_pack_resources.append(new_resource)
                    self.update(single_resources + new_pack_resources)
                    return

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        """Sets data in the model."""
        if role != Qt.ItemDataRole.CheckStateRole or not index.isValid():
            return False
        checked = value == Qt.CheckState.Checked.value
        self.checked_state_changed.emit(index, checked)
        return True

    def set_initial_state(self, selected_items):
        """Fills model with incomplete data; needs a call to :func:`update` to make the model usable.

        Args:
            selected_items (dict): mapping from item label to checked flag
        """
        for label, selected in selected_items.items():
            self._pack_resources.append(self.PackItem(label, [], selected))

    def set_checked(self, index, checked):
        """Checks or unchecks given item.

        Args:
            index (QModelIndex): resource label
            checked (bool): checked flag
        """
        row = index.row()
        if row < len(self._single_resources):
            self._single_resources[row] = self._single_resources[row]._replace(checked=checked)
        else:
            row -= len(self._single_resources)
            self._pack_resources[row] = self._pack_resources[row]._replace(checked=checked)
        self.dataChanged.emit(index, index, [Qt.ItemDataRole.CheckStateRole])

    def index_with_file_path(self):
        """Tries to find an item that has a valid file path or URL.

        Returns:
            QModelIndex: index to an item with usable file path/URL, or an invalid index if none found
        """
        for row, item in enumerate(self._single_resources):
            if item.checked and item.resource.url:
                return self.index(row, 0)
        for pack_row, item in enumerate(self._pack_resources):
            if item.checked:
                for row, resource in enumerate(item.resources):
                    if resource.hasfilepath:
                        return self.index(row, 0, self.index(pack_row, 0))
        return QModelIndex()


class ToolCommandLineArgsModel(CommandLineArgsModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._spec_args_root = CommandLineArgItem("Specification arguments")
        self._tool_args_root = CommandLineArgItem("Tool arguments", drop_enabled=True)
        self.appendRow(self._spec_args_root)
        self.appendRow(self._tool_args_root)
        self._tool_args_root.appendRow(NewCommandLineArgItem())

    def reset_model(self, spec_args, tool_args):
        self._args = tool_args
        self._reset_root(self._spec_args_root, spec_args, {}, has_empty_row=False)
        self._reset_root(
            self._tool_args_root,
            tool_args,
            {"editable": True, "selectable": True, "drag_enabled": True},
            has_empty_row=True,
        )

    def canDropMimeData(self, data, drop_action, row, column, parent):
        return parent.data() is not None and row >= 0



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains OptionsWidget class."""
import functools
from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QSpinBox,
    QStyle,
    QTableWidget,
    QToolButton,
    QWidget,
)
from ..commands import SetReaderOption


class OptionsWidget(QTableWidget):
    """A widget for handling simple options."""

    options_changed = Signal(dict)
    """Emitted whenever an option in the widget is changed."""
    about_to_undo = Signal(str)
    """Emitted before undo action."""
    load_default_mapping_requested = Signal()
    """Emitted whenever the uses presses the button to Load default mapping."""

    def __init__(self, undo_stack):
        """
        Args:
            undo_stack (QUndoStack): undo stack
        """
        super().__init__()
        self._options = {}
        self._connector = None
        self._undo_stack = undo_stack
        self._undo_enabled = True
        self._current_source_table = None
        self._ui_choices = {str: QLineEdit, list: QComboBox, int: QSpinBox, bool: QCheckBox}
        self._ui_elements = {}
        self._build_ui()
        self.horizontalHeader().setVisible(False)
        self.verticalHeader().setVisible(False)
        self.setShowGrid(False)

    def _clear_ui(self):
        """Clears UI."""
        self._ui_elements.clear()
        self.clear()
        self.setRowCount(0)
        self.setColumnCount(0)

    def _build_ui(self):
        """Builds ui from specification in dict"""
        self.insertRow(0)
        key_options = sorted(self._options.items(), key=lambda x: list(self._ui_choices).index(x[1]["type"]))
        for column, (key, options) in enumerate(key_options):
            ui_element = self._ui_choices[options["type"]]()
            maximum = options.get("Maximum", None)
            if maximum is not None:
                ui_element.setMaximum(maximum)
            minimum = options.get("Minimum", None)
            if minimum is not None:
                ui_element.setMinimum(minimum)
            special_value_text = options.get("SpecialValueText", None)
            if special_value_text is not None:
                ui_element.setSpecialValueText(special_value_text)
            max_length = options.get("MaxLength", None)
            if max_length is not None:
                ui_element.setMaxLength(max_length)
            bound_arguments = {"option_key": key, "options_widget": self}
            if isinstance(ui_element, QSpinBox):
                handler = functools.partial(_emit_spin_box_option_changed, **bound_arguments)
                ui_element.valueChanged.connect(handler)
            elif isinstance(ui_element, QLineEdit):
                handler = functools.partial(_emit_line_edit_option_changed, **bound_arguments)
                ui_element.textChanged.connect(handler)
                ui_element.setMaximumWidth(40)
            elif isinstance(ui_element, QCheckBox):
                handler = functools.partial(_emit_check_box_option_changed, **bound_arguments)
                ui_element.stateChanged.connect(handler)
            elif isinstance(ui_element, QComboBox):
                ui_element.addItems([str(x) for x in options["Items"]])
                handler = functools.partial(_emit_combo_box_option_changed, **bound_arguments)
                ui_element.currentTextChanged.connect(handler)
            self._ui_elements[key] = ui_element
            # Add to layout:
            widget = QWidget()
            layout = QHBoxLayout(widget)
            layout.setContentsMargins(3, 3, 3, 3)
            layout.setSpacing(3)
            layout.addWidget(QLabel(options["label"] + ":"))
            layout.addWidget(ui_element)
            self.insertColumn(self.columnCount())
            self.setCellWidget(0, column, widget)
        self.resizeColumnsToContents()
        self.resizeRowsToContents()
        self._adjust_height()

    def _adjust_height(self):
        height = self.verticalHeader().length() + 2
        if self.horizontalScrollBar().isVisible():
            height += self.style().pixelMetric(QStyle.PixelMetric.PM_ScrollBarExtent)
        self.setMaximumHeight(height)

    def resizeEvent(self, ev):
        super().resizeEvent(ev)
        self._adjust_height()

    @property
    def connector(self):
        """The connection manager linked to this options widget."""
        return self._connector

    def set_connector(self, connector):
        """Sets connection manager.

        Args:
            connector (ConnectionManager): connector
        """
        self._connector = connector
        self._options = connector.connection.BASE_OPTIONS.copy()
        self._options.update(connector.connection.OPTIONS)
        connector.current_table_changed.connect(self._fetch_options_from_connector)
        self.options_changed.connect(connector.update_options)
        self._clear_ui()
        self._build_ui()
        self._set_options(self._connector.current_table)
        if hasattr(self._connector.connection, "create_default_mapping"):
            button = QToolButton()
            button.setText("Load default mapping")
            self.insertColumn(self.columnCount())
            self.setCellWidget(0, self.columnCount() - 1, button)
            button.clicked.connect(self.load_default_mapping_requested)

    @property
    def undo_stack(self):
        return self._undo_stack

    @property
    def undo_enabled(self):
        return self._undo_enabled

    @property
    def current_source_table(self):
        return self._current_source_table

    def _set_options(self, source_table, options=None):
        """Sets state of options

        Args:
            source_table (str): name of the source table
            options (dict, optional): Dict with option name as key and value as value (default: {None})
        """
        self._current_source_table = source_table
        if options is None:
            options = {}
        for key, ui_element in self._ui_elements.items():
            default = self._options[key]["default"]
            value = options.get(key, default)
            if value is None:
                continue
            ui_element.blockSignals(True)
            if isinstance(ui_element, QSpinBox):
                ui_element.setValue(value)
            elif isinstance(ui_element, QLineEdit):
                ui_element.setText(value)
            elif isinstance(ui_element, QCheckBox):
                ui_element.setChecked(value)
            elif isinstance(ui_element, QComboBox):
                ui_element.setCurrentText(value)
            ui_element.blockSignals(False)

    def set_option_without_undo(self, source_table, option_key, value):
        self.about_to_undo.emit(source_table)
        ui_element = self._ui_elements[option_key]
        if isinstance(ui_element, QSpinBox):
            current_value = ui_element.value()
            if value == current_value:
                return
            self._undo_enabled = False
            ui_element.setValue(value)
            self._undo_enabled = True
        elif isinstance(ui_element, QLineEdit):
            current_value = ui_element.text()
            if value == current_value:
                return
            self._undo_enabled = False
            ui_element.setText(value)
            self._undo_enabled = True
        elif isinstance(ui_element, QCheckBox):
            current_value = ui_element.isChecked()
            if value == current_value:
                return
            self._undo_enabled = False
            ui_element.setChecked(value)
            self._undo_enabled = True
        elif isinstance(ui_element, QComboBox):
            current_value = ui_element.currentText()
            if value == current_value:
                return
            self._undo_enabled = False
            ui_element.setCurrentText(value)
            self._undo_enabled = True

    @Slot()
    def _fetch_options_from_connector(self):
        """Read options from the connector."""
        table_name = self._connector.current_table
        self._set_options(table_name, self._connector.get_current_options())


def _emit_spin_box_option_changed(i, option_key, options_widget):
    """
    A 'slot' to transform changes in QSpinBox into changes in options.

    Args:
        i (int): spin box value
        option_key (str): option's key
        options_widget (OptionsWidget): options widget
    """
    if options_widget.undo_enabled:
        previous_value = options_widget.connector.get_current_option_value(option_key)
        options_widget.undo_stack.push(
            SetReaderOption(options_widget.current_source_table, option_key, options_widget, i, previous_value)
        )
    options = {option_key: i}
    options_widget.options_changed.emit(options)


def _emit_line_edit_option_changed(text, option_key, options_widget):
    """
    A 'slot' to transform changes in QLineEdit into changes in options.

    Args:
        text (str): text for undo/redo
        option_key (str): option's key
        options_widget (OptionsWidget): options widget
    """
    if options_widget.undo_enabled:
        previous_value = options_widget.connector.get_current_option_value(option_key)
        options_widget.undo_stack.push(
            SetReaderOption(options_widget.current_source_table, option_key, options_widget, text, previous_value)
        )
    options = {option_key: text}
    options_widget.options_changed.emit(options)


def _emit_check_box_option_changed(state, option_key, options_widget):
    """
    A 'slot' to transform changes in QCheckBox into changes in options.

    Args:
        state (int): check box value
        option_key (str): option's key
        options_widget (OptionsWidget): options widget
    """
    checked = state == Qt.CheckState.Checked.value
    if options_widget.undo_enabled:
        previous_value = options_widget.connector.get_current_option_value(option_key)
        options_widget.undo_stack.push(
            SetReaderOption(options_widget.current_source_table, option_key, options_widget, checked, previous_value)
        )
    options = {option_key: checked}
    options_widget.options_changed.emit(options)


def _emit_combo_box_option_changed(text, option_key, options_widget):
    """
    A 'slot' to transform changes in QComboBox into changes in options.

    Args:
        text (str): text for undo/redo
        option_key (str): option's key
        options_widget (OptionsWidget): options widget
    """
    if options_widget.undo_enabled:
        previous_value = options_widget.connector.get_current_option_value(option_key)
        options_widget.undo_stack.push(
            SetReaderOption(options_widget.current_source_table, option_key, options_widget, text, previous_value)
        )
    options = {option_key: text}
    options_widget.options_changed.emit(options)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""View properties widget."""
from PySide6.QtCore import QPoint, Slot
from spinetoolbox.widgets.properties_widget import PropertiesWidgetBase
from .custom_menus import ViewRefsContextMenu, ViewSelectionsContextMenu


class ViewPropertiesWidget(PropertiesWidgetBase):
    """Widget for the View Item Properties."""

    def __init__(self, toolbox):
        """
        Args:
            toolbox (ToolboxUI): The toolbox instance where this widget should be embedded
        """
        from ..ui.view_properties import Ui_Form  # pylint: disable=import-outside-toplevel

        super().__init__(toolbox)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        # Class attributes
        self.view_prop_context_menu = None
        self.connect_signals()

    def connect_signals(self):
        """Connect signals to slots."""
        self.ui.treeView_references.customContextMenuRequested.connect(self.show_view_refs_context_menu)
        self.ui.treeView_pinned_values.customContextMenuRequested.connect(self.show_view_selections_context_menu)

    @Slot(QPoint)
    def show_view_refs_context_menu(self, pos):
        """Create and show a context-menu in View refs.

        Args:
            pos (QPoint): Mouse position
        """
        view = self._toolbox.active_project_item
        global_pos = self.ui.treeView_references.viewport().mapToGlobal(pos)
        self.view_prop_context_menu = ViewRefsContextMenu(self, global_pos, view)
        option = self.view_prop_context_menu.get_action()
        if option == "Pin values...":
            view.pin_values()
        elif option == "Open editor...":
            view.open_editor()
        self.view_prop_context_menu.deleteLater()
        self.view_prop_context_menu = None

    @Slot(QPoint)
    def show_view_selections_context_menu(self, pos):
        """Create and show a context-menu in View selections.

        Args:
            pos (QPoint): Mouse position
        """
        view = self._toolbox.active_project_item
        global_pos = self.ui.treeView_pinned_values.viewport().mapToGlobal(pos)
        self.view_prop_context_menu = ViewSelectionsContextMenu(self, global_pos, view)
        option = self.view_prop_context_menu.get_action()
        if option == "Plot":
            view.plot_selected_pinned_values()
        elif option == "Copy plot data":
            view.copy_selected_pinned_value_plot_data()
        elif option == "Unpin":
            view.unpin_selected_pinned_values()
        elif option == "Rename...":
            view.renamed_selected_pinned_value()
        self.view_prop_context_menu.deleteLater()
        self.view_prop_context_menu = None



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for the ``output_channel`` module."""
from pathlib import Path
import unittest
from spine_items.exporter.output_channel import OutputChannel


class TestOutputChannel(unittest.TestCase):
    def test_initialization_with_out_label(self):
        channel = OutputChannel("In label", "Exporter 1", "Out label")
        self.assertEqual(channel.in_label, "In label")
        self.assertEqual(channel.out_label, "Out label")

    def test_initialization_without_out_label(self):
        channel = OutputChannel("In label", "Exporter 1")
        self.assertEqual(channel.in_label, "In label")
        self.assertEqual(channel.out_label, "In label_exported@Exporter 1")

    def test_serialization(self):
        channel = OutputChannel("In label", "Exporter 1", "Out label")
        channel_dict = channel.to_dict("/project/path/")
        restored = OutputChannel.from_dict(channel_dict, "Exporter 1", "/project/path/")
        self.assertEqual(restored.in_label, "In label")
        self.assertEqual(restored.out_label, "Out label")

    def test_serialization_hides_credentials_from_url(self):
        channel = OutputChannel(
            "In label",
            "Exporter 1",
            "Out label",
            {
                "dialect": "mysql",
                "host": "dummy.test.org",
                "port": 99,
                "database": "data_leak",
                "username": "admin",
                "password": "s3cr3t",
            },
        )
        channel_dict = channel.to_dict("/project/path/")
        self.assertNotIn("username", channel_dict["out_url"])
        self.assertNotIn("password", channel_dict["out_url"])
        restored = OutputChannel.from_dict(channel_dict, "Exporter 1", "/project/path/")
        self.assertEqual(restored.in_label, "In label")
        self.assertEqual(restored.out_label, "Out label")
        self.assertEqual(
            restored.out_url, {"dialect": "mysql", "host": "dummy.test.org", "port": 99, "database": "data_leak"}
        )

    def test_database_paths_serialized_as_relative_when_inside_project_dir(self):
        initial_project_dir = Path("project_1", "dir")
        relative_database_path = Path("data", "base.sqlite")
        channel = OutputChannel(
            "In label",
            "Exporter 1",
            "Out label",
            {"dialect": "sqlite", "database": str(initial_project_dir / relative_database_path)},
        )
        channel_dict = channel.to_dict(str(initial_project_dir))
        self.assertNotIn("username", channel_dict["out_url"])
        self.assertNotIn("password", channel_dict["out_url"])
        new_project_dir = Path("project_2")
        restored = OutputChannel.from_dict(channel_dict, "Exporter 1", str(new_project_dir))
        self.assertEqual(restored.in_label, "In label")
        self.assertEqual(restored.out_label, "Out label")
        self.assertEqual(
            restored.out_url, {"dialect": "sqlite", "database": str(new_project_dir / relative_database_path)}
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

################################################################################
## Form generated from reading UI file 'tool_properties.ui'
##
## Created by: Qt User Interface Compiler version 6.7.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QCheckBox, QComboBox,
    QFrame, QGridLayout, QHBoxLayout, QHeaderView,
    QLabel, QLineEdit, QPushButton, QRadioButton,
    QScrollArea, QSizePolicy, QSpacerItem, QSplitter,
    QToolButton, QTreeView, QVBoxLayout, QWidget)

from ...widgets import ArgsTreeView
from spinetoolbox.widgets.custom_qwidgets import ElidedLabel
from spine_items import resources_icons_rc

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(312, 612)
        Form.setStyleSheet(u"QScrollArea { background: transparent; }\n"
"QScrollArea > QWidget > QWidget { background: transparent; }")
        self.verticalLayout = QVBoxLayout(Form)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.scrollArea = QScrollArea(Form)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setFrameShape(QFrame.Shape.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 312, 612))
        self.scrollAreaWidgetContents.setAutoFillBackground(False)
        self.verticalLayout_3 = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_9 = QHBoxLayout()
        self.horizontalLayout_9.setSpacing(6)
        self.horizontalLayout_9.setObjectName(u"horizontalLayout_9")
        self.horizontalLayout_9.setContentsMargins(-1, -1, -1, 6)
        self.label_tool_specification = QLabel(self.scrollAreaWidgetContents)
        self.label_tool_specification.setObjectName(u"label_tool_specification")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_tool_specification.sizePolicy().hasHeightForWidth())
        self.label_tool_specification.setSizePolicy(sizePolicy)
        self.label_tool_specification.setMaximumSize(QSize(16777215, 16777215))

        self.horizontalLayout_9.addWidget(self.label_tool_specification)

        self.comboBox_tool = QComboBox(self.scrollAreaWidgetContents)
        self.comboBox_tool.setObjectName(u"comboBox_tool")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.comboBox_tool.sizePolicy().hasHeightForWidth())
        self.comboBox_tool.setSizePolicy(sizePolicy1)

        self.horizontalLayout_9.addWidget(self.comboBox_tool)

        self.toolButton_tool_specification = QToolButton(self.scrollAreaWidgetContents)
        self.toolButton_tool_specification.setObjectName(u"toolButton_tool_specification")
        icon = QIcon()
        icon.addFile(u":/icons/wrench.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toolButton_tool_specification.setIcon(icon)
        self.toolButton_tool_specification.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)

        self.horizontalLayout_9.addWidget(self.toolButton_tool_specification)


        self.verticalLayout_3.addLayout(self.horizontalLayout_9)

        self.horizontalLayout_options = QHBoxLayout()
        self.horizontalLayout_options.setObjectName(u"horizontalLayout_options")
        self.horizontalLayout_options.setContentsMargins(-1, 0, -1, 6)

        self.verticalLayout_3.addLayout(self.horizontalLayout_options)

        self.splitter = QSplitter(self.scrollAreaWidgetContents)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Vertical)
        self.splitter.setChildrenCollapsible(False)
        self.treeView_cmdline_args = ArgsTreeView(self.splitter)
        self.treeView_cmdline_args.setObjectName(u"treeView_cmdline_args")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.treeView_cmdline_args.sizePolicy().hasHeightForWidth())
        self.treeView_cmdline_args.setSizePolicy(sizePolicy2)
        self.treeView_cmdline_args.setAcceptDrops(True)
        self.treeView_cmdline_args.setEditTriggers(QAbstractItemView.EditTrigger.AnyKeyPressed|QAbstractItemView.EditTrigger.DoubleClicked|QAbstractItemView.EditTrigger.EditKeyPressed)
        self.treeView_cmdline_args.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.treeView_cmdline_args.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.treeView_cmdline_args.setTextElideMode(Qt.TextElideMode.ElideLeft)
        self.splitter.addWidget(self.treeView_cmdline_args)
        self.treeView_cmdline_args.header().setMinimumSectionSize(26)
        self.gridLayoutWidget = QWidget(self.splitter)
        self.gridLayoutWidget.setObjectName(u"gridLayoutWidget")
        self.gridLayout = QGridLayout(self.gridLayoutWidget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer_2, 0, 1, 1, 1)

        self.toolButton_remove_arg = QToolButton(self.gridLayoutWidget)
        self.toolButton_remove_arg.setObjectName(u"toolButton_remove_arg")
        icon1 = QIcon()
        icon1.addFile(u":/icons/minus.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toolButton_remove_arg.setIcon(icon1)

        self.gridLayout.addWidget(self.toolButton_remove_arg, 0, 2, 1, 1)

        self.toolButton_add_file_path_arg = QToolButton(self.gridLayoutWidget)
        self.toolButton_add_file_path_arg.setObjectName(u"toolButton_add_file_path_arg")
        icon2 = QIcon()
        icon2.addFile(u":/icons/file-upload.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toolButton_add_file_path_arg.setIcon(icon2)
        self.toolButton_add_file_path_arg.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)

        self.gridLayout.addWidget(self.toolButton_add_file_path_arg, 0, 0, 1, 1)

        self.treeView_input_files = QTreeView(self.gridLayoutWidget)
        self.treeView_input_files.setObjectName(u"treeView_input_files")
        sizePolicy2.setHeightForWidth(self.treeView_input_files.sizePolicy().hasHeightForWidth())
        self.treeView_input_files.setSizePolicy(sizePolicy2)
        self.treeView_input_files.setDragEnabled(False)
        self.treeView_input_files.setDragDropMode(QAbstractItemView.DragDropMode.DragOnly)
        self.treeView_input_files.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.treeView_input_files.setTextElideMode(Qt.TextElideMode.ElideLeft)
        self.treeView_input_files.setUniformRowHeights(True)
        self.treeView_input_files.setAnimated(False)
        self.treeView_input_files.header().setMinimumSectionSize(26)

        self.gridLayout.addWidget(self.treeView_input_files, 1, 0, 1, 3)

        self.splitter.addWidget(self.gridLayoutWidget)

        self.verticalLayout_3.addWidget(self.splitter)

        self.frame = QFrame(self.scrollAreaWidgetContents)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.StyledPanel)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_2 = QVBoxLayout(self.frame)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_2 = QLabel(self.frame)
        self.label_2.setObjectName(u"label_2")

        self.horizontalLayout.addWidget(self.label_2)

        self.radioButton_execute_in_source = QRadioButton(self.frame)
        self.radioButton_execute_in_source.setObjectName(u"radioButton_execute_in_source")

        self.horizontalLayout.addWidget(self.radioButton_execute_in_source)

        self.radioButton_execute_in_work = QRadioButton(self.frame)
        self.radioButton_execute_in_work.setObjectName(u"radioButton_execute_in_work")
        self.radioButton_execute_in_work.setChecked(True)

        self.horizontalLayout.addWidget(self.radioButton_execute_in_work)


        self.verticalLayout_2.addLayout(self.horizontalLayout)

        self.label = QLabel(self.frame)
        self.label.setObjectName(u"label")

        self.verticalLayout_2.addWidget(self.label)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.lineEdit_root_directory = QLineEdit(self.frame)
        self.lineEdit_root_directory.setObjectName(u"lineEdit_root_directory")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.lineEdit_root_directory.sizePolicy().hasHeightForWidth())
        self.lineEdit_root_directory.setSizePolicy(sizePolicy3)
        self.lineEdit_root_directory.setClearButtonEnabled(True)

        self.horizontalLayout_3.addWidget(self.lineEdit_root_directory)

        self.toolButton_browse_root_directory = QToolButton(self.frame)
        self.toolButton_browse_root_directory.setObjectName(u"toolButton_browse_root_directory")
        icon3 = QIcon()
        icon3.addFile(u":/icons/folder-open-solid.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toolButton_browse_root_directory.setIcon(icon3)

        self.horizontalLayout_3.addWidget(self.toolButton_browse_root_directory)


        self.verticalLayout_2.addLayout(self.horizontalLayout_3)

        self.label_group_id = QLabel(self.frame)
        self.label_group_id.setObjectName(u"label_group_id")

        self.verticalLayout_2.addWidget(self.label_group_id)

        self.lineEdit_group_id = QLineEdit(self.frame)
        self.lineEdit_group_id.setObjectName(u"lineEdit_group_id")
        sizePolicy3.setHeightForWidth(self.lineEdit_group_id.sizePolicy().hasHeightForWidth())
        self.lineEdit_group_id.setSizePolicy(sizePolicy3)
        self.lineEdit_group_id.setClearButtonEnabled(True)

        self.verticalLayout_2.addWidget(self.lineEdit_group_id)

        self.label_jupyter = ElidedLabel(self.frame)
        self.label_jupyter.setObjectName(u"label_jupyter")

        self.verticalLayout_2.addWidget(self.label_jupyter)

        self.kill_consoles_check_box = QCheckBox(self.frame)
        self.kill_consoles_check_box.setObjectName(u"kill_consoles_check_box")

        self.verticalLayout_2.addWidget(self.kill_consoles_check_box)

        self.log_process_output_check_box = QCheckBox(self.frame)
        self.log_process_output_check_box.setObjectName(u"log_process_output_check_box")

        self.verticalLayout_2.addWidget(self.log_process_output_check_box)

        self.horizontalLayout_11 = QHBoxLayout()
        self.horizontalLayout_11.setSpacing(6)
        self.horizontalLayout_11.setObjectName(u"horizontalLayout_11")
        self.pushButton_tool_results = QPushButton(self.frame)
        self.pushButton_tool_results.setObjectName(u"pushButton_tool_results")
        sizePolicy3.setHeightForWidth(self.pushButton_tool_results.sizePolicy().hasHeightForWidth())
        self.pushButton_tool_results.setSizePolicy(sizePolicy3)

        self.horizontalLayout_11.addWidget(self.pushButton_tool_results)

        self.lineEdit_result_directory = QLineEdit(self.frame)
        self.lineEdit_result_directory.setObjectName(u"lineEdit_result_directory")
        self.lineEdit_result_directory.setReadOnly(False)
        self.lineEdit_result_directory.setClearButtonEnabled(True)

        self.horizontalLayout_11.addWidget(self.lineEdit_result_directory)

        self.toolButton_browse_result_directory = QToolButton(self.frame)
        self.toolButton_browse_result_directory.setObjectName(u"toolButton_browse_result_directory")
        self.toolButton_browse_result_directory.setIcon(icon3)

        self.horizontalLayout_11.addWidget(self.toolButton_browse_result_directory)


        self.verticalLayout_2.addLayout(self.horizontalLayout_11)


        self.verticalLayout_3.addWidget(self.frame)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.verticalLayout.addWidget(self.scrollArea)

        QWidget.setTabOrder(self.comboBox_tool, self.toolButton_tool_specification)
        QWidget.setTabOrder(self.toolButton_tool_specification, self.treeView_cmdline_args)
        QWidget.setTabOrder(self.treeView_cmdline_args, self.toolButton_add_file_path_arg)
        QWidget.setTabOrder(self.toolButton_add_file_path_arg, self.toolButton_remove_arg)
        QWidget.setTabOrder(self.toolButton_remove_arg, self.treeView_input_files)
        QWidget.setTabOrder(self.treeView_input_files, self.radioButton_execute_in_source)
        QWidget.setTabOrder(self.radioButton_execute_in_source, self.radioButton_execute_in_work)
        QWidget.setTabOrder(self.radioButton_execute_in_work, self.lineEdit_root_directory)
        QWidget.setTabOrder(self.lineEdit_root_directory, self.toolButton_browse_root_directory)
        QWidget.setTabOrder(self.toolButton_browse_root_directory, self.lineEdit_group_id)
        QWidget.setTabOrder(self.lineEdit_group_id, self.kill_consoles_check_box)
        QWidget.setTabOrder(self.kill_consoles_check_box, self.log_process_output_check_box)
        QWidget.setTabOrder(self.log_process_output_check_box, self.pushButton_tool_results)
        QWidget.setTabOrder(self.pushButton_tool_results, self.lineEdit_result_directory)
        QWidget.setTabOrder(self.lineEdit_result_directory, self.toolButton_browse_result_directory)
        QWidget.setTabOrder(self.toolButton_browse_result_directory, self.scrollArea)

        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.label_tool_specification.setText(QCoreApplication.translate("Form", u"Specification:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_tool.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Tool specification for this Tool</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.toolButton_tool_specification.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Tool specification options.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.toolButton_remove_arg.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Remove selected tool command line args</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.toolButton_remove_arg.setText(QCoreApplication.translate("Form", u"...", None))
#if QT_CONFIG(tooltip)
        self.toolButton_add_file_path_arg.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Append selected Available resources to Tool arguments</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.toolButton_add_file_path_arg.setText("")
        self.label_2.setText(QCoreApplication.translate("Form", u"Execute in", None))
        self.radioButton_execute_in_source.setText(QCoreApplication.translate("Form", u"Source dir", None))
        self.radioButton_execute_in_work.setText(QCoreApplication.translate("Form", u"Work dir", None))
        self.label.setText(QCoreApplication.translate("Form", u"Source code root directory:", None))
        self.lineEdit_root_directory.setText("")
        self.lineEdit_root_directory.setPlaceholderText(QCoreApplication.translate("Form", u"Root directory path...", None))
        self.label_group_id.setText(QCoreApplication.translate("Form", u"Reuse console id:", None))
#if QT_CONFIG(tooltip)
        self.lineEdit_group_id.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Enter an id for sharing a console with other Tools in this project. Leave empty to run this Tool in isolation.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_jupyter.setText(QCoreApplication.translate("Form", u"Console info", None))
#if QT_CONFIG(tooltip)
        self.kill_consoles_check_box.setToolTip(QCoreApplication.translate("Form", u"If checked, console processes will be killed automatically after execution finishes freeing memory and other resources.", None))
#endif // QT_CONFIG(tooltip)
        self.kill_consoles_check_box.setText(QCoreApplication.translate("Form", u"Kill consoles at the end of execution", None))
        self.log_process_output_check_box.setText(QCoreApplication.translate("Form", u"Log process output to a file", None))
#if QT_CONFIG(tooltip)
        self.pushButton_tool_results.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Open results archive in file browser</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_tool_results.setText(QCoreApplication.translate("Form", u"Results...", None))
#if QT_CONFIG(tooltip)
        self.lineEdit_result_directory.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Result directory</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.lineEdit_result_directory.setText("")
#if QT_CONFIG(tooltip)
        self.toolButton_browse_result_directory.setToolTip(QCoreApplication.translate("Form", u"<html><head/><body><p>Select Result directory using file browser. Leave blank to use default directory.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
    # retranslateUi




######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Undo/redo commands that can be used by multiple project items."""
from spinetoolbox.project_commands import SpineToolboxCommand


class UpdateCancelOnErrorCommand(SpineToolboxCommand):
    def __init__(self, item_name, cancel_on_error, project):
        """Command to update Importer, Exporter, and Merger cancel on error setting.

        Args:
            item_name (str): Item's name
            cancel_on_error (bool): New setting
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._item_name = item_name
        self._redo_cancel_on_error = cancel_on_error
        self._undo_cancel_on_error = not cancel_on_error
        self._project = project
        self.setText(f"change {item_name} cancel on error setting")

    def redo(self):
        item = self._project.get_item(self._item_name)
        item.set_cancel_on_error(self._redo_cancel_on_error)

    def undo(self):
        item = self._project.get_item(self._item_name)
        item.set_cancel_on_error(self._undo_cancel_on_error)


class UpdateOnConflictCommand(SpineToolboxCommand):
    def __init__(self, item_name, on_conflict, project):
        """Command to update Importer and Merger 'on conflict' setting.

        Args:
            item_name (str): Item's name
            on_conflict (str): New setting
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._item_name = item_name
        self._redo_on_conflict = on_conflict
        project_item = project.get_item(item_name)
        self._undo_on_conflict = project_item.on_conflict
        self._project = project
        self.setText(f"change {item_name} on conflict setting")

    def redo(self):
        item = self._project.get_item(self._item_name)
        item.set_on_conflict(self._redo_on_conflict)

    def undo(self):
        item = self._project.get_item(self._item_name)
        item.set_on_conflict(self._undo_on_conflict)


class ChangeItemSelectionCommand(SpineToolboxCommand):
    def __init__(self, item_name, model, index, selected):
        """Command to change file item's selection status.

        Args:
            item_name (str): project item's name
            model (FileListModel): File model
            index (QModelIndex): Index to file model
            selected (bool): True if the item is selected, False otherwise
        """
        super().__init__()
        self._model = model
        self._index = index
        self._selected = selected
        self.setText(f"change {item_name} file selection")

    def redo(self):
        self._model.set_checked(self._index, self._selected)

    def undo(self):
        self._model.set_checked(self._index, not self._selected)


class UpdateCmdLineArgsCommand(SpineToolboxCommand):
    def __init__(self, item_name, cmd_line_args, project):
        """Command to update Tool command line args.

        Args:
            item_name (str): item's name
            cmd_line_args (list): list of command line args
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._item_name = item_name
        self._redo_cmd_line_args = cmd_line_args
        item = project.get_item(item_name)
        self._undo_cmd_line_args = item.cmd_line_args
        self._project = project
        self.setText(f"change command line arguments of {item_name}")

    def redo(self):
        item = self._project.get_item(self._item_name)
        item.update_cmd_line_args(self._redo_cmd_line_args)

    def undo(self):
        item = self._project.get_item(self._item_name)
        item.update_cmd_line_args(self._undo_cmd_line_args)


class UpdateGroupIdCommand(SpineToolboxCommand):
    def __init__(self, item_name, group_id, project):
        """Command to update item group identifier.

        Args:
            item_name (str): item's name
            group_id (str): group identifier
            project (SpineToolboxProject): project
        """
        super().__init__()
        self._item_name = item_name
        self._redo_group_id = group_id
        item = project.get_item(item_name)
        self._undo_group_id = item.group_id
        self._project = project
        self.setText(f"change group identifier of {item_name}")

    def redo(self):
        item = self._project.get_item(self._item_name)
        item.do_set_group_id(self._redo_group_id)

    def undo(self):
        item = self._project.get_item(self._item_name)
        item.do_set_group_id(self._undo_group_id)


class UpdateRootDirCommand(SpineToolboxCommand):
    def __init__(self, item_name, root_dir, project):
        """Command to update Tool root directory.

        Args:
            item_name (str): Item's name
            root_dir (str): Root directory
            project (SpineToolboxProject): Project
        """
        super().__init__()
        self._item_name = item_name
        self._redo_root_dir = root_dir
        item = project.get_item(item_name)
        self._undo_root_dir = item.root_dir
        self._project = project
        self.setText(f"change root directory of {item_name}")

    def redo(self):
        item = self._project.get_item(self._item_name)
        item.do_set_root_directory(self._redo_root_dir)

    def undo(self):
        item = self._project.get_item(self._item_name)
        item.do_set_root_directory(self._undo_root_dir)


class UpdateResultDirCommand(SpineToolboxCommand):
    def __init__(self, item_name, result_dir, project):
        """Command to update Tool result directory.

        Args:
            item_name (str): Item's name
            result_dir (str): Result directory
            project (SpineToolboxProject): Project
        """
        super().__init__()
        self._item_name = item_name
        self._redo_result_dir = result_dir
        item = project.get_item(item_name)
        self._undo_result_dir = item._options.get("output_directory")
        self._project = project
        self.setText(f"change result directory of {item_name}")

    def redo(self):
        item = self._project.get_item(self._item_name)
        item.do_set_result_directory(self._redo_result_dir)

    def undo(self):
        item = self._project.get_item(self._item_name)
        item.do_set_result_directory(self._undo_result_dir)
